import concurrent.futures as cf
import sys
from random import random
from unittest import mock

import pytest
from UnleashClient import UnleashClient

import sentry_sdk
from sentry_sdk import start_span, start_transaction
from sentry_sdk.integrations.unleash import UnleashIntegration
from tests.conftest import ApproxDict
from tests.integrations.unleash.testutils import mock_unleash_client


def test_is_enabled(sentry_init, capture_events, uninstall_integration):
    uninstall_integration(UnleashIntegration.identifier)

    with mock_unleash_client():
        client = UnleashClient()  # type: ignore[arg-type]
        sentry_init(integrations=[UnleashIntegration()])
        client.is_enabled("hello")
        client.is_enabled("world")
        client.is_enabled("other")

    events = capture_events()
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 1
    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": False},
            {"flag": "other", "result": False},
        ]
    }


def test_is_enabled_threaded(sentry_init, capture_events, uninstall_integration):
    uninstall_integration(UnleashIntegration.identifier)

    with mock_unleash_client():
        client = UnleashClient()  # type: ignore[arg-type]
        sentry_init(integrations=[UnleashIntegration()])
        events = capture_events()

        def task(flag_key):
            # Creates a new isolation scope for the thread.
            # This means the evaluations in each task are captured separately.
            with sentry_sdk.isolation_scope():
                client.is_enabled(flag_key)
                # use a tag to identify to identify events later on
                sentry_sdk.set_tag("task_id", flag_key)
                sentry_sdk.capture_exception(Exception("something wrong!"))

        # Capture an eval before we split isolation scopes.
        client.is_enabled("hello")

        with cf.ThreadPoolExecutor(max_workers=2) as pool:
            pool.map(task, ["world", "other"])

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": False},
        ]
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7 or higher")
def test_is_enabled_asyncio(sentry_init, capture_events, uninstall_integration):
    asyncio = pytest.importorskip("asyncio")
    uninstall_integration(UnleashIntegration.identifier)

    with mock_unleash_client():
        client = UnleashClient()  # type: ignore[arg-type]
        sentry_init(integrations=[UnleashIntegration()])
        events = capture_events()

        async def task(flag_key):
            with sentry_sdk.isolation_scope():
                client.is_enabled(flag_key)
                # use a tag to identify to identify events later on
                sentry_sdk.set_tag("task_id", flag_key)
                sentry_sdk.capture_exception(Exception("something wrong!"))

        async def runner():
            return asyncio.gather(task("world"), task("other"))

        # Capture an eval before we split isolation scopes.
        client.is_enabled("hello")

        asyncio.run(runner())

    # Capture error in original scope
    sentry_sdk.set_tag("task_id", "0")
    sentry_sdk.capture_exception(Exception("something wrong!"))

    assert len(events) == 3
    events.sort(key=lambda e: e["tags"]["task_id"])

    assert events[0]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
        ]
    }
    assert events[1]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "other", "result": False},
        ]
    }
    assert events[2]["contexts"]["flags"] == {
        "values": [
            {"flag": "hello", "result": True},
            {"flag": "world", "result": False},
        ]
    }


def test_wraps_original(sentry_init, uninstall_integration):
    with mock_unleash_client():
        client = UnleashClient()  # type: ignore[arg-type]

        mock_is_enabled = mock.Mock(return_value=random() < 0.5)
        client.is_enabled = mock_is_enabled

        uninstall_integration(UnleashIntegration.identifier)
        sentry_init(integrations=[UnleashIntegration()])  # type: ignore

    res = client.is_enabled("test-flag", "arg", kwarg=1)
    assert res == mock_is_enabled.return_value
    assert mock_is_enabled.call_args == (
        ("test-flag", "arg"),
        {"kwarg": 1},
    )


def test_wrapper_attributes(sentry_init, uninstall_integration):
    with mock_unleash_client():
        client = UnleashClient()  # type: ignore[arg-type]

        original_is_enabled = client.is_enabled

        uninstall_integration(UnleashIntegration.identifier)
        sentry_init(integrations=[UnleashIntegration()])  # type: ignore

        # Mock clients methods have not lost their qualified names after decoration.
        assert client.is_enabled.__name__ == "is_enabled"
        assert client.is_enabled.__qualname__ == original_is_enabled.__qualname__


@pytest.mark.parametrize(
    "span_streaming",
    [True, False],
)
def test_unleash_span_integration(
    sentry_init, capture_events, capture_items, uninstall_integration, span_streaming
):
    uninstall_integration(UnleashIntegration.identifier)

    with mock_unleash_client():
        sentry_init(
            traces_sample_rate=1.0,
            integrations=[UnleashIntegration()],
            trace_lifecycle="stream" if span_streaming else "static",
        )

        client = UnleashClient()  # type: ignore[arg-type]

        if span_streaming:
            items = capture_items("span")
            with sentry_sdk.traces.start_span(name="bar"):
                client.is_enabled("hello")
                client.is_enabled("other")

            sentry_sdk.flush()

            assert len(items) == 1
            span = items[0].payload
            assert span["attributes"]["flag.evaluation.hello"] is True
            assert span["attributes"]["flag.evaluation.other"] is False

        else:
            events = capture_events()
            with start_transaction(name="hi"):
                with start_span(op="foo", name="bar"):
                    client.is_enabled("hello")
                    client.is_enabled("other")

            (event,) = events
            assert event["spans"][0]["data"] == ApproxDict(
                {"flag.evaluation.hello": True, "flag.evaluation.other": False}
            )



import subprocess
import sys
from textwrap import dedent

import pytest
from typer.testing import CliRunner

runner = CliRunner()


def test_catch_exceptions(tmpdir):
    app = tmpdir.join("app.py")

    app.write(
        dedent(
            """
    import typer
    from unittest import mock

    from sentry_sdk import init, transport
    from sentry_sdk.integrations.typer import TyperIntegration

    def capture_envelope(self, envelope):
        print("capture_envelope was called")
        event = envelope.get_event()
        if event is not None:
            print(event)

    transport.HttpTransport.capture_envelope = capture_envelope

    init("http://foobar@localhost/123", integrations=[TyperIntegration()])

    app = typer.Typer()

    @app.command()
    def test():
        print("test called")
        raise Exception("pollo")

    app()
    """
        )
    )

    with pytest.raises(subprocess.CalledProcessError) as excinfo:
        subprocess.check_output([sys.executable, str(app)], stderr=subprocess.STDOUT)

    output = excinfo.value.output

    assert b"capture_envelope was called" in output
    assert b"test called" in output
    assert b"pollo" in output



import inspect

import dill
import pytest
from apache_beam.runners.common import DoFnContext, DoFnInvoker
from apache_beam.transforms.core import CallableWrapperDoFn, DoFn, ParDo, _DoFnParam
from apache_beam.typehints.decorators import getcallargs_forhints
from apache_beam.typehints.trivial_inference import instance_to_type
from apache_beam.utils.windowed_value import WindowedValue

from sentry_sdk.integrations.beam import (
    BeamIntegration,
    _wrap_inspect_call,
    _wrap_task_call,
)

try:
    from apache_beam.runners.common import OutputHandler
except ImportError:
    from apache_beam.runners.common import OutputProcessor as OutputHandler


def foo():
    return True


def bar(x, y):
    # print(x + y)
    return True


def baz(x, y=2):
    # print(x + y)
    return True


class A:
    def __init__(self, fn):
        self.r = "We are in A"
        self.fn = fn
        self._inspect_fn = _wrap_inspect_call(self, "fn")

    def process(self):
        return self.fn()


class B(A):
    def fa(self, x, element=False, another_element=False):
        if x or (element and not another_element):
            # print(self.r)
            return True
        1 / 0
        return False

    def __init__(self):
        self.r = "We are in B"
        super().__init__(self.fa)


class SimpleFunc(DoFn):
    def process(self, x):
        if x:
            1 / 0
        return [True]


class PlaceHolderFunc(DoFn):
    def process(self, x, timestamp=DoFn.TimestampParam, wx=DoFn.WindowParam):
        if isinstance(timestamp, _DoFnParam) or isinstance(wx, _DoFnParam):
            raise Exception("Bad instance")
        if x:
            1 / 0
        yield True


def fail(x):
    if x:
        1 / 0
    return [True]


test_parent = A(foo)
test_child = B()
test_simple = SimpleFunc()
test_place_holder = PlaceHolderFunc()
test_callable = CallableWrapperDoFn(fail)


# Cannot call simple functions or placeholder test.
@pytest.mark.parametrize(
    "obj,f,args,kwargs",
    [
        [test_parent, "fn", (), {}],
        [test_child, "fn", (False,), {"element": True}],
        [test_child, "fn", (True,), {}],
        [test_simple, "process", (False,), {}],
        [test_callable, "process", (False,), {}],
    ],
)
def test_monkey_patch_call(obj, f, args, kwargs):
    func = getattr(obj, f)

    assert func(*args, **kwargs)
    assert _wrap_task_call(func)(*args, **kwargs)


@pytest.mark.parametrize("f", [foo, bar, baz, test_parent.fn, test_child.fn])
def test_monkey_patch_pickle(f):
    f_temp = _wrap_task_call(f)
    assert dill.pickles(f_temp), "{} is not pickling correctly!".format(f)

    # Pickle everything
    s1 = dill.dumps(f_temp)
    s2 = dill.loads(s1)
    dill.dumps(s2)


@pytest.mark.parametrize(
    "f,args,kwargs",
    [
        [foo, (), {}],
        [bar, (1, 5), {}],
        [baz, (1,), {}],
        [test_parent.fn, (), {}],
        [test_child.fn, (False,), {"element": True}],
        [test_child.fn, (True,), {}],
    ],
)
def test_monkey_patch_signature(f, args, kwargs):
    arg_types = [instance_to_type(v) for v in args]
    kwargs_types = {k: instance_to_type(v) for (k, v) in kwargs.items()}
    f_temp = _wrap_task_call(f)
    try:
        getcallargs_forhints(f, *arg_types, **kwargs_types)
    except Exception:
        print("Failed on {} with parameters {}, {}".format(f, args, kwargs))
        raise
    try:
        getcallargs_forhints(f_temp, *arg_types, **kwargs_types)
    except Exception:
        print("Failed on {} with parameters {}, {}".format(f_temp, args, kwargs))
        raise
    try:
        expected_signature = inspect.signature(f)
        test_signature = inspect.signature(f_temp)
        assert expected_signature == test_signature, (
            "Failed on {}, signature {} does not match {}".format(
                f, expected_signature, test_signature
            )
        )
    except Exception:
        # expected to pass for py2.7
        pass


class _OutputHandler(OutputHandler):
    def process_outputs(
        self, windowed_input_element, results, watermark_estimator=None
    ):
        self.handle_process_outputs(
            windowed_input_element, results, watermark_estimator
        )

    def handle_process_outputs(
        self, windowed_input_element, results, watermark_estimator=None
    ):
        print(windowed_input_element)
        try:
            for result in results:
                assert result
        except StopIteration:
            print("In here")


@pytest.fixture
def init_beam(sentry_init):
    def inner(fn):
        sentry_init(default_integrations=False, integrations=[BeamIntegration()])
        # Little hack to avoid having to run the whole pipeline.
        pardo = ParDo(fn)
        signature = pardo._signature
        output_processor = _OutputHandler()
        return DoFnInvoker.create_invoker(
            signature,
            output_processor,
            DoFnContext("test"),
            input_args=[],
            input_kwargs={},
        )

    return inner


@pytest.mark.parametrize("fn", [test_simple, test_callable, test_place_holder])
def test_invoker_normal(init_beam, fn):
    invoker = init_beam(fn)
    print("Normal testing {} with {} invoker.".format(fn, invoker))
    windowed_value = WindowedValue(False, 0, [None])
    invoker.invoke_process(windowed_value)


@pytest.mark.parametrize("fn", [test_simple, test_callable, test_place_holder])
def test_invoker_exception(init_beam, capture_events, capture_exceptions, fn):
    invoker = init_beam(fn)
    events = capture_events()

    print("Exception testing {} with {} invoker.".format(fn, invoker))
    # Window value will always have one value for the process to run.
    windowed_value = WindowedValue(True, 0, [None])
    try:
        invoker.invoke_process(windowed_value)
    except Exception:
        pass

    (event,) = events
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    assert exception["mechanism"]["type"] == "beam"



import uuid

import dramatiq
import pytest
from dramatiq.brokers.stub import StubBroker
from dramatiq.middleware import Middleware, SkipMessage

import sentry_sdk
from sentry_sdk import start_transaction
from sentry_sdk.consts import SPANDATA, SPANSTATUS
from sentry_sdk.integrations.dramatiq import DramatiqIntegration
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.tracing import Transaction, TransactionSource

ignore_logger("dramatiq.worker.WorkerThread")


@pytest.fixture(scope="function")
def broker(request, sentry_init):
    param = getattr(request, "param", None)
    if isinstance(param, dict):
        sentry_init(integrations=[DramatiqIntegration()], **param)
    else:
        sentry_init(
            integrations=[DramatiqIntegration()],
            traces_sample_rate=param,
        )
    broker = StubBroker()
    broker.emit_after("process_boot")
    dramatiq.set_broker(broker)
    yield broker
    broker.flush_all()
    broker.close()


@pytest.fixture
def worker(broker):
    worker = dramatiq.Worker(broker, worker_timeout=100, worker_threads=1)
    worker.start()
    yield worker
    worker.stop()


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_a_single_error_is_captured(broker, worker, capture_events, fail_fast):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        return x / y

    dummy_actor.send(1, 2)
    dummy_actor.send(1, 0)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    (event,) = events
    exception = event["exception"]["values"][0]
    assert exception["type"] == "ZeroDivisionError"


@pytest.mark.parametrize(
    "broker,expected_span_status,fail_fast,span_streaming",
    [
        ({"traces_sample_rate": 1.0}, SPANSTATUS.INTERNAL_ERROR, False, False),
        ({"traces_sample_rate": 1.0}, SPANSTATUS.OK, False, False),
        ({"traces_sample_rate": 1.0}, SPANSTATUS.INTERNAL_ERROR, True, False),
        ({"traces_sample_rate": 1.0}, SPANSTATUS.OK, True, False),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            SPANSTATUS.INTERNAL_ERROR,
            False,
            True,
        ),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            SPANSTATUS.OK,
            False,
            True,
        ),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            SPANSTATUS.INTERNAL_ERROR,
            True,
            True,
        ),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            SPANSTATUS.OK,
            True,
            True,
        ),
    ],
    ids=[
        "error",
        "success",
        "error_fail_fast",
        "success_fail_fast",
        "error_stream",
        "success_stream",
        "error_fail_fast_stream",
        "success_fail_fast_stream",
    ],
    indirect=["broker"],
)
def test_task_transaction(
    broker,
    worker,
    capture_events,
    capture_items,
    expected_span_status,
    fail_fast,
    span_streaming,
):
    task_fails = expected_span_status == SPANSTATUS.INTERNAL_ERROR

    if span_streaming:
        items = capture_items("event", "span")
    else:
        events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        return x / y

    dummy_actor.send(1, int(not task_fails))

    if expected_span_status == SPANSTATUS.INTERNAL_ERROR and fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)

    worker.join()
    sentry_sdk.flush()

    if span_streaming:
        if task_fails:
            error_item, segment_item = items
            error_event = error_item.payload
            exception = error_event["exception"]["values"][0]
            assert exception["type"] == "ZeroDivisionError"
            assert exception["mechanism"]["type"] == DramatiqIntegration.identifier
        else:
            (segment_item,) = items

        segment = segment_item.payload
        assert segment_item.type == "span"
        assert segment["name"] == "dummy_actor"
        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.op"] == "queue.task.dramatiq"
        assert segment["attributes"]["sentry.segment.name.source"] == "task"
        assert (
            segment["attributes"][SPANDATA.MESSAGING_DESTINATION_NAME]
            == dummy_actor.queue_name
        )
        assert segment["status"] == ("error" if task_fails else "ok")
    else:
        if task_fails:
            error_event = events.pop(0)
            exception = error_event["exception"]["values"][0]
            assert exception["type"] == "ZeroDivisionError"
            assert exception["mechanism"]["type"] == DramatiqIntegration.identifier

        (event,) = events
        assert event["type"] == "transaction"
        assert event["transaction"] == "dummy_actor"
        assert event["transaction_info"] == {"source": TransactionSource.TASK}
        assert (
            event["contexts"]["trace"]["data"][SPANDATA.MESSAGING_DESTINATION_NAME]
            == dummy_actor.queue_name
        )
        assert event["contexts"]["trace"]["status"] == expected_span_status


@pytest.mark.parametrize(
    "broker,span_streaming",
    [
        ({"traces_sample_rate": 1.0}, False),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            True,
        ),
    ],
    ids=["static", "stream"],
    indirect=["broker"],
)
def test_dramatiq_propagate_trace(
    broker, worker, capture_events, capture_items, span_streaming
):
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="outer") as outer_span:

            @dramatiq.actor(max_retries=0)
            def propagated_trace_task():
                pass

            propagated_trace_task.send()
            broker.join(propagated_trace_task.queue_name)
            worker.join()

        sentry_sdk.flush()

        inner_segment, outer_segment = [i.payload for i in items]
        assert inner_segment["name"] == "propagated_trace_task"
        assert inner_segment["attributes"]["sentry.op"] == "queue.task.dramatiq"
        assert inner_segment["trace_id"] == outer_span.trace_id
        assert outer_segment["name"] == "outer"
    else:
        events = capture_events()

        @dramatiq.actor(max_retries=0)
        def propagated_trace_task():
            pass

        with start_transaction() as outer_transaction:
            propagated_trace_task.send()
            broker.join(propagated_trace_task.queue_name)
            worker.join()

        assert (
            events[0]["transaction"] == "propagated_trace_task"
        )  # the "inner" transaction
        assert events[0]["contexts"]["trace"]["trace_id"] == outer_transaction.trace_id


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_dramatiq_message_id_is_set_as_extra(
    broker, worker, capture_events, fail_fast
):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        sentry_sdk.capture_message("hi")
        return x / y

    dummy_actor.send(1, 0)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    event_message, event_error = events
    assert "dramatiq_message_id" in event_message["extra"]
    assert "dramatiq_message_id" in event_error["extra"]
    assert (
        event_message["extra"]["dramatiq_message_id"]
        == event_error["extra"]["dramatiq_message_id"]
    )
    msg_ids = [e["extra"]["dramatiq_message_id"] for e in events]
    assert all(uuid.UUID(msg_id) and isinstance(msg_id, str) for msg_id in msg_ids)


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_local_variables_are_captured(broker, worker, capture_events, fail_fast):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        foo = 42  # noqa
        return x / y

    dummy_actor.send(1, 2)
    dummy_actor.send(1, 0)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    (event,) = events
    exception = event["exception"]["values"][0]
    assert exception["stacktrace"]["frames"][-1]["vars"] == {
        "x": "1",
        "y": "0",
        "foo": "42",
    }


def test_that_messages_are_captured(broker, worker, capture_events):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor():
        sentry_sdk.capture_message("hi")

    dummy_actor.send()
    broker.join(dummy_actor.queue_name)
    worker.join()

    (event,) = events
    assert event["message"] == "hi"
    assert event["level"] == "info"
    assert event["transaction"] == "dummy_actor"


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_sub_actor_errors_are_captured(broker, worker, capture_events, fail_fast):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        sub_actor.send(x, y)

    @dramatiq.actor(max_retries=0)
    def sub_actor(x, y):
        return x / y

    dummy_actor.send(1, 2)
    dummy_actor.send(1, 0)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    (event,) = events
    assert event["transaction"] == "sub_actor"

    exception = event["exception"]["values"][0]
    assert exception["type"] == "ZeroDivisionError"


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_multiple_errors_are_captured(broker, worker, capture_events, fail_fast):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        return x / y

    dummy_actor.send(1, 0)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    dummy_actor.send(1, None)
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    event1, event2 = events

    assert event1["transaction"] == "dummy_actor"
    exception = event1["exception"]["values"][0]
    assert exception["type"] == "ZeroDivisionError"

    assert event2["transaction"] == "dummy_actor"
    exception = event2["exception"]["values"][0]
    assert exception["type"] == "TypeError"


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_message_data_is_added_as_request(
    broker, worker, capture_events, fail_fast
):
    events = capture_events()

    @dramatiq.actor(max_retries=0)
    def dummy_actor(x, y):
        return x / y

    dummy_actor.send_with_options(
        args=(
            1,
            0,
        ),
        max_retries=0,
    )
    if fail_fast:
        with pytest.raises(ZeroDivisionError):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    (event,) = events

    assert event["transaction"] == "dummy_actor"
    request_data = event["contexts"]["dramatiq"]["data"]
    assert request_data["queue_name"] == "default"
    assert request_data["actor_name"] == "dummy_actor"
    assert request_data["args"] == [1, 0]
    assert request_data["kwargs"] == {}
    assert request_data["options"]["max_retries"] == 0
    assert uuid.UUID(request_data["message_id"])
    assert isinstance(request_data["message_timestamp"], int)


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_expected_exceptions_are_not_captured(
    broker, worker, capture_events, fail_fast
):
    events = capture_events()

    class ExpectedException(Exception):
        pass

    @dramatiq.actor(max_retries=0, throws=ExpectedException)
    def dummy_actor():
        raise ExpectedException

    dummy_actor.send()
    if fail_fast:
        with pytest.raises(ExpectedException):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    assert events == []


@pytest.mark.parametrize(
    "fail_fast",
    [
        False,
        True,
    ],
)
def test_that_retry_exceptions_are_not_captured(
    broker, worker, capture_events, fail_fast
):
    events = capture_events()

    @dramatiq.actor(max_retries=2)
    def dummy_actor():
        raise dramatiq.errors.Retry("Retrying", delay=100)

    dummy_actor.send()
    if fail_fast:
        with pytest.raises(dramatiq.errors.Retry):
            broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    else:
        broker.join(dummy_actor.queue_name, fail_fast=fail_fast)
    worker.join()

    assert events == []


@pytest.mark.parametrize(
    "broker,span_streaming",
    [
        ({"traces_sample_rate": 1.0}, False),
        (
            {
                "traces_sample_rate": 1.0,
                "trace_lifecycle": "stream",
            },
            True,
        ),
    ],
    ids=["static", "stream"],
    indirect=["broker"],
)
def test_that_skip_message_cleans_up_scope_and_transaction(
    broker, worker, capture_events, capture_items, span_streaming
):
    captured_spans: list = []

    class SkipMessageMiddleware(Middleware):
        def before_process_message(self, broker, message):
            if span_streaming:
                captured_spans.append(sentry_sdk.get_current_span())
            else:
                captured_spans.append(sentry_sdk.get_current_scope().transaction)
            raise SkipMessage()

    broker.add_middleware(SkipMessageMiddleware())

    if span_streaming:
        items = capture_items("span")

    @dramatiq.actor(max_retries=0)
    def skipped_actor(): ...

    skipped_actor.send()

    broker.join(skipped_actor.queue_name)
    worker.join()

    if span_streaming:
        sentry_sdk.flush()
        (segment_payload,) = [i.payload for i in items]
        assert segment_payload["name"] == "skipped_actor"
        assert segment_payload["end_timestamp"] is not None
    else:
        (transaction,) = captured_spans
        assert isinstance(transaction, Transaction)
        assert transaction.timestamp is not None



import pytest
import responses
from gql import Client, __version__, gql
from gql.transport.exceptions import TransportQueryError
from gql.transport.requests import RequestsHTTPTransport

from sentry_sdk.integrations.gql import GQLIntegration
from sentry_sdk.utils import parse_version

GQL_VERSION = parse_version(__version__)


@responses.activate
def _execute_mock_query(response_json):
    url = "http://example.com/graphql"
    query_string = """
        query Example {
            example
        }
    """

    # Mock the GraphQL server response
    responses.add(
        method=responses.POST,
        url=url,
        json=response_json,
        status=200,
    )

    transport = RequestsHTTPTransport(url=url)
    client = Client(transport=transport)
    query = gql(query_string)

    return client.execute(query)


@responses.activate
def _execute_mock_query_with_keyword_document(response_json):
    url = "http://example.com/graphql"
    query_string = """
        query Example {
            example
        }
    """

    # Mock the GraphQL server response
    responses.add(
        method=responses.POST,
        url=url,
        json=response_json,
        status=200,
    )

    transport = RequestsHTTPTransport(url=url)
    client = Client(transport=transport)
    query = gql(query_string)

    return client.execute(document=query)


_execute_query_funcs = [_execute_mock_query]
if GQL_VERSION < (4,):
    _execute_query_funcs.append(_execute_mock_query_with_keyword_document)


def _make_erroneous_query(capture_events, execute_query):
    """
    Make an erroneous GraphQL query, and assert that the error was reraised, that
    exactly one event was recorded, and that the exception recorded was a
    TransportQueryError. Then, return the event to allow further verifications.
    """
    events = capture_events()
    response_json = {"errors": ["something bad happened"]}

    with pytest.raises(TransportQueryError):
        execute_query(response_json)

    assert len(events) == 1, (
        "the sdk captured %d events, but 1 event was expected" % len(events)
    )

    (event,) = events
    (exception,) = event["exception"]["values"]

    assert exception["type"] == "TransportQueryError", (
        "%s was captured, but we expected a TransportQueryError" % exception(type)
    )

    assert "request" in event

    return event


def test_gql_init(sentry_init):
    """
    Integration test to ensure we can initialize the SDK with the GQL Integration
    """
    sentry_init(integrations=[GQLIntegration()])


@pytest.mark.parametrize("execute_query", _execute_query_funcs)
def test_real_gql_request_no_error(sentry_init, capture_events, execute_query):
    """
    Integration test verifying that the GQLIntegration works as expected with successful query.
    """
    sentry_init(integrations=[GQLIntegration()])
    events = capture_events()

    response_data = {"example": "This is the example"}
    response_json = {"data": response_data}

    result = execute_query(response_json)

    assert result == response_data, (
        "client.execute returned a different value from what it received from the server"
    )
    assert len(events) == 0, (
        "the sdk captured an event, even though the query was successful"
    )


@pytest.mark.parametrize("execute_query", _execute_query_funcs)
def test_real_gql_request_with_error_no_pii(sentry_init, capture_events, execute_query):
    """
    Integration test verifying that the GQLIntegration works as expected with query resulting
    in a GraphQL error, and that PII is not sent.
    """
    sentry_init(integrations=[GQLIntegration()])

    event = _make_erroneous_query(capture_events, execute_query)

    assert "data" not in event["request"]
    assert "response" not in event["contexts"]


@pytest.mark.parametrize("execute_query", _execute_query_funcs)
def test_real_gql_request_with_error_with_pii(
    sentry_init, capture_events, execute_query
):
    """
    Integration test verifying that the GQLIntegration works as expected with query resulting
    in a GraphQL error, and that PII is not sent.
    """
    sentry_init(integrations=[GQLIntegration()], send_default_pii=True)

    event = _make_erroneous_query(capture_events, execute_query)

    assert "data" in event["request"]
    assert "response" in event["contexts"]



import pytest
from mockupdb import MockupDB, OpQuery
from pymongo import MongoClient

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.pymongo import PyMongoIntegration, _strip_pii


@pytest.fixture(scope="session")
def mongo_server():
    server = MockupDB(verbose=True)
    server.autoresponds("ismaster", maxWireVersion=8)
    server.run()
    server.autoresponds(
        {"find": "test_collection"}, cursor={"id": 123, "firstBatch": []}
    )
    # Find query changed somewhere between PyMongo 3.1 and 3.12.
    # This line is to respond to "find" queries sent by old PyMongo the same way it's done above.
    server.autoresponds(OpQuery({"foobar": 1}), cursor={"id": 123, "firstBatch": []})
    server.autoresponds({"insert": "test_collection"}, ok=1)
    server.autoresponds({"insert": "erroneous"}, ok=0, errmsg="test error")
    yield server
    server.stop()


@pytest.mark.parametrize("with_pii", [False, True])
def test_transactions(sentry_init, capture_events, mongo_server, with_pii):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=with_pii,
    )
    events = capture_events()

    connection = MongoClient(mongo_server.uri)

    with start_transaction():
        list(
            connection["test_db"]["test_collection"].find({"foobar": 1})
        )  # force query execution
        connection["test_db"]["test_collection"].insert_one({"foo": 2})
        try:
            connection["test_db"]["erroneous"].insert_many([{"bar": 3}, {"baz": 4}])
            pytest.fail("Request should raise")
        except Exception:
            pass

    (event,) = events
    (find, insert_success, insert_fail) = event["spans"]

    common_tags = {
        "db.name": "test_db",
        "db.system": "mongodb",
        "db.driver.name": "pymongo",
        "net.peer.name": mongo_server.host,
        "net.peer.port": str(mongo_server.port),
    }
    for span in find, insert_success, insert_fail:
        assert span["data"][SPANDATA.DB_SYSTEM] == "mongodb"
        assert span["data"][SPANDATA.DB_DRIVER_NAME] == "pymongo"
        assert span["data"][SPANDATA.DB_NAME] == "test_db"
        assert span["data"][SPANDATA.SERVER_ADDRESS] == "localhost"
        assert span["data"][SPANDATA.SERVER_PORT] == mongo_server.port
        for field, value in common_tags.items():
            assert span["tags"][field] == value
            assert span["data"][field] == value

    assert find["op"] == "db"
    assert insert_success["op"] == "db"
    assert insert_fail["op"] == "db"

    assert find["data"]["db.operation"] == "find"
    assert find["tags"]["db.operation"] == "find"
    assert insert_success["data"]["db.operation"] == "insert"
    assert insert_success["tags"]["db.operation"] == "insert"
    assert insert_fail["data"]["db.operation"] == "insert"
    assert insert_fail["tags"]["db.operation"] == "insert"

    assert find["description"].startswith('{"find')
    assert insert_success["description"].startswith('{"insert')
    assert insert_fail["description"].startswith('{"insert')

    assert find["data"][SPANDATA.DB_MONGODB_COLLECTION] == "test_collection"
    assert find["tags"][SPANDATA.DB_MONGODB_COLLECTION] == "test_collection"
    assert insert_success["data"][SPANDATA.DB_MONGODB_COLLECTION] == "test_collection"
    assert insert_success["tags"][SPANDATA.DB_MONGODB_COLLECTION] == "test_collection"
    assert insert_fail["data"][SPANDATA.DB_MONGODB_COLLECTION] == "erroneous"
    assert insert_fail["tags"][SPANDATA.DB_MONGODB_COLLECTION] == "erroneous"
    if with_pii:
        assert "1" in find["description"]
        assert "2" in insert_success["description"]
        assert "3" in insert_fail["description"] and "4" in insert_fail["description"]
    else:
        # All values in filter replaced by "%s"
        assert "1" not in find["description"]
        # All keys below top level replaced by "%s"
        assert "2" not in insert_success["description"]
        assert (
            "3" not in insert_fail["description"]
            and "4" not in insert_fail["description"]
        )

    assert find["status"] == "ok"
    assert find["tags"]["status"] == "ok"
    assert insert_success["status"] == "ok"
    assert insert_success["tags"]["status"] == "ok"
    assert insert_fail["status"] == "internal_error"
    assert insert_fail["tags"]["status"] == "internal_error"


@pytest.mark.parametrize("with_pii", [False, True])
def test_segment_span_streaming(sentry_init, capture_items, mongo_server, with_pii):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=with_pii,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    connection = MongoClient(mongo_server.uri)

    with sentry_sdk.traces.start_span(name="test_segment"):
        list(
            connection["test_db"]["test_collection"].find({"foobar": 1})
        )  # force query execution
        connection["test_db"]["test_collection"].insert_one({"foo": 2})
        try:
            connection["test_db"]["erroneous"].insert_many([{"bar": 3}, {"baz": 4}])
            pytest.fail("Request should raise")
        except Exception:
            pass
    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 4

    (find, insert_success, insert_fail, segment) = spans
    assert segment["name"] == "test_segment"

    for span in find, insert_success, insert_fail:
        attrs = span["attributes"]
        assert attrs["db.system.name"] == "mongodb"
        assert attrs[SPANDATA.DB_DRIVER_NAME] == "pymongo"
        assert attrs["db.namespace"] == "test_db"
        assert attrs[SPANDATA.SERVER_ADDRESS] == "localhost"
        assert attrs[SPANDATA.SERVER_PORT] == mongo_server.port
        assert attrs["sentry.op"] == "db"
        assert attrs["sentry.origin"] == "auto.db.pymongo"

    assert find["attributes"]["db.operation.name"] == "find"
    assert insert_success["attributes"]["db.operation.name"] == "insert"
    assert insert_fail["attributes"]["db.operation.name"] == "insert"

    assert find["name"].startswith('{"find')
    assert insert_success["name"].startswith('{"insert')
    assert insert_fail["name"].startswith('{"insert')

    assert find["attributes"]["db.collection.name"] == "test_collection"
    assert insert_success["attributes"]["db.collection.name"] == "test_collection"
    assert insert_fail["attributes"]["db.collection.name"] == "erroneous"

    for span in find, insert_success, insert_fail:
        assert span["attributes"][SPANDATA.DB_QUERY_TEXT] == span["name"]

    if with_pii:
        assert "1" in find["name"]
        assert "2" in insert_success["name"]
        assert "3" in insert_fail["name"] and "4" in insert_fail["name"]
    else:
        assert "1" not in find["name"]
        assert "2" not in insert_success["name"]
        assert "3" not in insert_fail["name"] and "4" not in insert_fail["name"]

    assert find["status"] == "ok"
    assert insert_success["status"] == "ok"
    assert insert_fail["status"] == "error"


@pytest.mark.parametrize("with_pii", [False, True])
def test_breadcrumbs(sentry_init, capture_events, mongo_server, with_pii):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=with_pii,
    )
    events = capture_events()

    connection = MongoClient(mongo_server.uri)

    list(
        connection["test_db"]["test_collection"].find({"foobar": 1})
    )  # force query execution
    capture_message("hi")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb["category"] == "query"
    assert crumb["message"].startswith('{"find')
    if with_pii:
        assert "1" in crumb["message"]
    else:
        assert "1" not in crumb["message"]
    assert crumb["type"] == "db"
    assert crumb["data"] == {
        "db.name": "test_db",
        "db.system": "mongodb",
        "db.driver.name": "pymongo",
        "db.operation": "find",
        "net.peer.name": mongo_server.host,
        "net.peer.port": str(mongo_server.port),
        "db.mongodb.collection": "test_collection",
    }


@pytest.mark.parametrize("with_pii", [False, True])
def test_breadcrumbs_span_streaming(sentry_init, capture_items, mongo_server, with_pii):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=with_pii,
        trace_lifecycle="stream",
    )
    items = capture_items("event")

    connection = MongoClient(mongo_server.uri)

    list(
        connection["test_db"]["test_collection"].find({"foobar": 1})
    )  # force query execution
    capture_message("hi")

    event = items[0].payload
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb["category"] == "query"
    assert crumb["message"].startswith('{"find')
    if with_pii:
        assert "1" in crumb["message"]
    else:
        assert "1" not in crumb["message"]
    assert crumb["type"] == "db"

    data = crumb["data"]
    assert data["db.namespace"] == "test_db"
    assert data["db.system.name"] == "mongodb"
    assert data["db.driver.name"] == "pymongo"
    assert data["db.operation.name"] == "find"
    assert data["db.collection.name"] == "test_collection"
    assert data["db.query.text"] == crumb["message"]
    assert data["sentry.op"] == "db"
    assert data["sentry.origin"] == "auto.db.pymongo"
    assert data[SPANDATA.SERVER_ADDRESS] == "localhost"
    assert data[SPANDATA.SERVER_PORT] == mongo_server.port


@pytest.mark.parametrize(
    "testcase",
    [
        {
            "command": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {
                        "username": "anton2",
                        "email": "anton@somewhere.io",
                        "password": "c4e86722fb56d946f7ddeecdae47e1c4458bf98a0a3ee5d5113111adf7bf0175",
                        "_id": "635bc7403cb4f8a736f61cf2",
                    }
                ],
            },
            "command_stripped": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {"username": "%s", "email": "%s", "password": "%s", "_id": "%s"}
                ],
            },
        },
        {
            "command": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {
                        "username": "indiana4",
                        "email": "indy@jones.org",
                        "password": "63e86722fb56d946f7ddeecdae47e1c4458bf98a0a3ee5d5113111adf7bf016b",
                        "_id": "635bc7403cb4f8a736f61cf3",
                    }
                ],
            },
            "command_stripped": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {"username": "%s", "email": "%s", "password": "%s", "_id": "%s"}
                ],
            },
        },
        {
            "command": {
                "find": "my_collection",
                "filter": {},
                "limit": 1,
                "singleBatch": True,
            },
            "command_stripped": {
                "find": "my_collection",
                "filter": {},
                "limit": 1,
                "singleBatch": True,
            },
        },
        {
            "command": {
                "find": "my_collection",
                "filter": {"username": "notthere"},
                "limit": 1,
                "singleBatch": True,
            },
            "command_stripped": {
                "find": "my_collection",
                "filter": {"username": "%s"},
                "limit": 1,
                "singleBatch": True,
            },
        },
        {
            "command": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {
                        "username": "userx1",
                        "email": "x@somewhere.io",
                        "password": "ccc86722fb56d946f7ddeecdae47e1c4458bf98a0a3ee5d5113111adf7bf0175",
                        "_id": "635bc7403cb4f8a736f61cf4",
                    },
                    {
                        "username": "userx2",
                        "email": "x@somewhere.io",
                        "password": "xxx86722fb56d946f7ddeecdae47e1c4458bf98a0a3ee5d5113111adf7bf0175",
                        "_id": "635bc7403cb4f8a736f61cf5",
                    },
                ],
            },
            "command_stripped": {
                "insert": "my_collection",
                "ordered": True,
                "documents": [
                    {"username": "%s", "email": "%s", "password": "%s", "_id": "%s"},
                    {"username": "%s", "email": "%s", "password": "%s", "_id": "%s"},
                ],
            },
        },
        {
            "command": {
                "find": "my_collection",
                "filter": {"email": "ada@lovelace.com"},
            },
            "command_stripped": {"find": "my_collection", "filter": {"email": "%s"}},
        },
        {
            "command": {
                "aggregate": "my_collection",
                "pipeline": [{"$match": {}}, {"$group": {"_id": 1, "n": {"$sum": 1}}}],
                "cursor": {},
            },
            "command_stripped": {
                "aggregate": "my_collection",
                "pipeline": [{"$match": {}}, {"$group": {"_id": 1, "n": {"$sum": 1}}}],
                "cursor": "%s",
            },
        },
        {
            "command": {
                "aggregate": "my_collection",
                "pipeline": [
                    {"$match": {"email": "x@somewhere.io"}},
                    {"$group": {"_id": 1, "n": {"$sum": 1}}},
                ],
                "cursor": {},
            },
            "command_stripped": {
                "aggregate": "my_collection",
                "pipeline": [
                    {"$match": {"email": "%s"}},
                    {"$group": {"_id": 1, "n": {"$sum": 1}}},
                ],
                "cursor": "%s",
            },
        },
        {
            "command": {
                "createIndexes": "my_collection",
                "indexes": [{"name": "username_1", "key": [("username", 1)]}],
            },
            "command_stripped": {
                "createIndexes": "my_collection",
                "indexes": [{"name": "username_1", "key": [("username", 1)]}],
            },
        },
        {
            "command": {
                "update": "my_collection",
                "ordered": True,
                "updates": [
                    ("q", {"email": "anton@somewhere.io"}),
                    (
                        "u",
                        {
                            "email": "anton2@somwehre.io",
                            "extra_field": "extra_content",
                            "new": "bla",
                        },
                    ),
                    ("multi", False),
                    ("upsert", False),
                ],
            },
            "command_stripped": {
                "update": "my_collection",
                "ordered": True,
                "updates": "%s",
            },
        },
        {
            "command": {
                "update": "my_collection",
                "ordered": True,
                "updates": [
                    ("q", {"email": "anton2@somwehre.io"}),
                    ("u", {"$rename": {"new": "new_field"}}),
                    ("multi", False),
                    ("upsert", False),
                ],
            },
            "command_stripped": {
                "update": "my_collection",
                "ordered": True,
                "updates": "%s",
            },
        },
        {
            "command": {
                "update": "my_collection",
                "ordered": True,
                "updates": [
                    ("q", {"email": "x@somewhere.io"}),
                    ("u", {"$rename": {"password": "pwd"}}),
                    ("multi", True),
                    ("upsert", False),
                ],
            },
            "command_stripped": {
                "update": "my_collection",
                "ordered": True,
                "updates": "%s",
            },
        },
        {
            "command": {
                "delete": "my_collection",
                "ordered": True,
                "deletes": [("q", {"username": "userx2"}), ("limit", 1)],
            },
            "command_stripped": {
                "delete": "my_collection",
                "ordered": True,
                "deletes": "%s",
            },
        },
        {
            "command": {
                "delete": "my_collection",
                "ordered": True,
                "deletes": [("q", {"email": "xplus@somewhere.io"}), ("limit", 0)],
            },
            "command_stripped": {
                "delete": "my_collection",
                "ordered": True,
                "deletes": "%s",
            },
        },
        {
            "command": {
                "findAndModify": "my_collection",
                "query": {"email": "ada@lovelace.com"},
                "new": False,
                "remove": True,
            },
            "command_stripped": {
                "findAndModify": "my_collection",
                "query": {"email": "%s"},
                "new": "%s",
                "remove": "%s",
            },
        },
        {
            "command": {
                "findAndModify": "my_collection",
                "query": {"email": "anton2@somewhere.io"},
                "new": False,
                "update": {"email": "anton3@somwehre.io", "extra_field": "xxx"},
                "upsert": False,
            },
            "command_stripped": {
                "findAndModify": "my_collection",
                "query": {"email": "%s"},
                "new": "%s",
                "update": {"email": "%s", "extra_field": "%s"},
                "upsert": "%s",
            },
        },
        {
            "command": {
                "findAndModify": "my_collection",
                "query": {"email": "anton3@somewhere.io"},
                "new": False,
                "update": {"$rename": {"extra_field": "extra_field2"}},
                "upsert": False,
            },
            "command_stripped": {
                "findAndModify": "my_collection",
                "query": {"email": "%s"},
                "new": "%s",
                "update": {"$rename": "%s"},
                "upsert": "%s",
            },
        },
        {
            "command": {
                "renameCollection": "test.my_collection",
                "to": "test.new_collection",
            },
            "command_stripped": {
                "renameCollection": "test.my_collection",
                "to": "test.new_collection",
            },
        },
        {
            "command": {"drop": "new_collection"},
            "command_stripped": {"drop": "new_collection"},
        },
    ],
)
def test_strip_pii(testcase):
    assert _strip_pii(testcase["command"]) == testcase["command_stripped"]


def test_span_origin(sentry_init, capture_events, mongo_server):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    connection = MongoClient(mongo_server.uri)

    with start_transaction():
        list(
            connection["test_db"]["test_collection"].find({"foobar": 1})
        )  # force query execution

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "manual"
    assert event["spans"][0]["origin"] == "auto.db.pymongo"


def test_span_origin_span_streaming(sentry_init, capture_items, mongo_server):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    connection = MongoClient(mongo_server.uri)

    with sentry_sdk.traces.start_span(name="test_segment"):
        list(
            connection["test_db"]["test_collection"].find({"foobar": 1})
        )  # force query execution
    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 2
    (db_span, segment) = spans
    assert segment["name"] == "test_segment"
    assert db_span["attributes"]["sentry.origin"] == "auto.db.pymongo"


def test_span_streaming_status_on_success(sentry_init, capture_items, mongo_server):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    connection = MongoClient(mongo_server.uri)

    with sentry_sdk.traces.start_span(name="test_segment"):
        connection["test_db"]["test_collection"].insert_one({"foo": 1})
    sentry_sdk.flush()

    spans = [item.payload for item in items]
    assert len(spans) == 2
    (db_span, segment) = spans
    assert segment["name"] == "test_segment"
    assert db_span["status"] == "ok"


def test_span_streaming_status_on_failure(sentry_init, capture_items, mongo_server):
    sentry_init(
        integrations=[PyMongoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream",
    )
    items = capture_items("span")

    connection = MongoClient(mongo_server.uri)

    with sentry_sdk.traces.start_span(name="test_segment"):
        try:
            connection["test_db"]["erroneous"].insert_many([{"bar": 3}])
            pytest.fail("Request should raise")
        except Exception:
            pass
    sentry_sdk.flush()

    spans = [item.payload for item in items]

    assert len(spans) == 2
    (db_span, segment) = spans

    assert segment["name"] == "test_segment"
    assert db_span["status"] == "error"



import logging
import warnings
from unittest import mock

import pytest

from sentry_sdk import get_client
from sentry_sdk.consts import VERSION
from sentry_sdk.integrations.logging import (
    LoggingIntegration,
    ignore_logger,
    ignore_logger_for_sentry_logs,
    unignore_logger,
    unignore_logger_for_sentry_logs,
)

other_logger = logging.getLogger("testfoo")
logger = logging.getLogger(__name__)


@pytest.fixture(autouse=True)
def reset_level():
    other_logger.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)


@pytest.mark.parametrize("logger", [logger, other_logger])
def test_logging_works_with_many_loggers(sentry_init, capture_events, logger):
    sentry_init(integrations=[LoggingIntegration(event_level="ERROR")])
    events = capture_events()

    logger.info("bread")
    logger.critical("LOL")
    (event,) = events
    assert event["level"] == "fatal"
    assert not event["logentry"]["params"]
    assert event["logentry"]["message"] == "LOL"
    assert event["logentry"]["formatted"] == "LOL"
    assert any(crumb["message"] == "bread" for crumb in event["breadcrumbs"]["values"])


@pytest.mark.parametrize("integrations", [None, [], [LoggingIntegration()]])
@pytest.mark.parametrize(
    "kwargs", [{"exc_info": None}, {}, {"exc_info": 0}, {"exc_info": False}]
)
def test_logging_defaults(integrations, sentry_init, capture_events, kwargs):
    sentry_init(integrations=integrations)
    events = capture_events()

    logger.info("bread")
    logger.critical("LOL", **kwargs)
    (event,) = events

    assert event["level"] == "fatal"
    assert any(crumb["message"] == "bread" for crumb in event["breadcrumbs"]["values"])
    assert not any(
        crumb["message"] == "LOL" for crumb in event["breadcrumbs"]["values"]
    )
    assert "threads" not in event


def test_logging_extra_data(sentry_init, capture_events):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.info("bread", extra=dict(foo=42))
    logger.critical("lol", extra=dict(bar=69))

    (event,) = events

    assert event["level"] == "fatal"
    assert event["extra"] == {"bar": 69}
    assert any(
        crumb["message"] == "bread" and crumb["data"] == {"foo": 42}
        for crumb in event["breadcrumbs"]["values"]
    )


def test_logging_extra_data_integer_keys(sentry_init, capture_events):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.critical("integer in extra keys", extra={1: 1})

    (event,) = events

    assert event["extra"] == {"1": 1}


@pytest.mark.parametrize(
    "enable_stack_trace_kwarg",
    (
        pytest.param({"exc_info": True}, id="exc_info"),
        pytest.param({"stack_info": True}, id="stack_info"),
    ),
)
def test_logging_stack_trace(sentry_init, capture_events, enable_stack_trace_kwarg):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.error("first", **enable_stack_trace_kwarg)
    logger.error("second")

    (
        event_with,
        event_without,
    ) = events

    assert event_with["level"] == "error"
    assert event_with["threads"]["values"][0]["stacktrace"]["frames"]

    assert event_without["level"] == "error"
    assert "threads" not in event_without


def test_logging_level(sentry_init, capture_events):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.setLevel(logging.WARNING)
    logger.error("hi")
    (event,) = events
    assert event["level"] == "error"
    assert event["logentry"]["message"] == "hi"
    assert event["logentry"]["formatted"] == "hi"

    del events[:]

    logger.setLevel(logging.ERROR)
    logger.warning("hi")
    assert not events


def test_custom_log_level_names(sentry_init, capture_events):
    levels = {
        logging.DEBUG: "debug",
        logging.INFO: "info",
        logging.WARN: "warning",
        logging.WARNING: "warning",
        logging.ERROR: "error",
        logging.CRITICAL: "fatal",
        logging.FATAL: "fatal",
    }

    # set custom log level names
    logging.addLevelName(logging.DEBUG, "custom level debüg: ")
    logging.addLevelName(logging.INFO, "")
    logging.addLevelName(logging.WARN, "custom level warn: ")
    logging.addLevelName(logging.WARNING, "custom level warning: ")
    logging.addLevelName(logging.ERROR, None)
    logging.addLevelName(logging.CRITICAL, "custom level critical: ")
    logging.addLevelName(logging.FATAL, "custom level 🔥: ")

    for logging_level, sentry_level in levels.items():
        logger.setLevel(logging_level)
        sentry_init(
            integrations=[LoggingIntegration(event_level=logging_level)],
            default_integrations=False,
        )
        events = capture_events()

        logger.log(logging_level, "Trying level %s", logging_level)
        assert events
        assert events[0]["level"] == sentry_level
        assert events[0]["logentry"]["message"] == "Trying level %s"
        assert events[0]["logentry"]["formatted"] == f"Trying level {logging_level}"
        assert events[0]["logentry"]["params"] == [logging_level]

        del events[:]


def test_logging_filters(sentry_init, capture_events):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    should_log = False

    class MyFilter(logging.Filter):
        def filter(self, record):
            return should_log

    logger.addFilter(MyFilter())
    logger.error("hi")

    assert not events

    should_log = True
    logger.error("hi")

    (event,) = events
    assert event["logentry"]["message"] == "hi"
    assert event["logentry"]["formatted"] == "hi"


def test_logging_captured_warnings(sentry_init, capture_events, recwarn):
    sentry_init(
        integrations=[LoggingIntegration(event_level="WARNING")],
        default_integrations=False,
    )
    events = capture_events()

    logging.captureWarnings(True)
    warnings.warn("first", stacklevel=2)
    warnings.warn("second", stacklevel=2)
    logging.captureWarnings(False)

    warnings.warn("third", stacklevel=2)

    assert len(events) == 2

    assert events[0]["level"] == "warning"
    # Captured warnings start with the path where the warning was raised
    assert "UserWarning: first" in events[0]["logentry"]["message"]
    assert "UserWarning: first" in events[0]["logentry"]["formatted"]
    # For warnings, the message and formatted message are the same
    assert events[0]["logentry"]["message"] == events[0]["logentry"]["formatted"]
    assert events[0]["logentry"]["params"] == []

    assert events[1]["level"] == "warning"
    assert "UserWarning: second" in events[1]["logentry"]["message"]
    assert "UserWarning: second" in events[1]["logentry"]["formatted"]
    # For warnings, the message and formatted message are the same
    assert events[1]["logentry"]["message"] == events[1]["logentry"]["formatted"]
    assert events[1]["logentry"]["params"] == []

    # "third" is emitted after captureWarnings(False), so it is not captured
    # by Sentry and instead surfaces as a normal warning. Using recwarn also
    # suppresses it in the test output. Filter by message rather than asserting
    # on the total count, since unrelated warnings from other tests can leak in.
    third_warnings = [w for w in recwarn if str(w.message) == "third"]
    assert len(third_warnings) == 1


def test_ignore_logger(sentry_init, capture_events, request):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    ignore_logger("testfoo")
    request.addfinalizer(lambda: unignore_logger("testfoo"))

    other_logger.error("hi")

    assert not events


def test_ignore_logger_whitespace_padding(sentry_init, capture_events, request):
    """Here we test insensitivity to whitespace padding of ignored loggers"""
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    ignore_logger("testfoo")
    request.addfinalizer(lambda: unignore_logger("testfoo"))

    padded_logger = logging.getLogger("       testfoo   ")
    padded_logger.error("hi")
    assert not events


def test_ignore_logger_wildcard(sentry_init, capture_events, request):
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    ignore_logger("testfoo.*")
    request.addfinalizer(lambda: unignore_logger("testfoo.*"))

    nested_logger = logging.getLogger("testfoo.submodule")

    logger.error("hi")

    nested_logger.error("bye")

    (event,) = events
    assert event["logentry"]["message"] == "hi"
    assert event["logentry"]["formatted"] == "hi"


def test_ignore_logger_does_not_affect_sentry_logs(sentry_init, capture_items, request):
    """ignore_logger should suppress events/breadcrumbs but not Sentry Logs."""
    sentry_init(enable_logs=True)
    items = capture_items("log")

    ignore_logger("testfoo")
    request.addfinalizer(lambda: unignore_logger("testfoo"))

    other_logger.error("hi")
    get_client().flush()

    logs = [item.payload for item in items]
    assert len(logs) == 1
    assert logs[0]["body"] == "hi"


def test_ignore_logger_for_sentry_logs(
    sentry_init, capture_envelopes, capture_items, request
):
    """ignore_logger_for_sentry_logs should suppress Sentry Logs but not events."""
    sentry_init(enable_logs=True)
    envelopes = capture_envelopes()
    items = capture_items("log")

    ignore_logger_for_sentry_logs("testfoo")
    request.addfinalizer(lambda: unignore_logger_for_sentry_logs("testfoo"))

    other_logger.error("hi")
    get_client().flush()

    # Event should still be captured
    event_envelopes = [e for e in envelopes if e.items[0].type == "event"]
    assert len(event_envelopes) == 1

    # But no Sentry Logs
    logs = [item.payload for item in items]
    assert len(logs) == 0


def test_logging_dictionary_interpolation(sentry_init, capture_events):
    """Here we test an entire dictionary being interpolated into the log message."""
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.error("this is a log with a dictionary %s", {"foo": "bar"})

    (event,) = events
    assert event["logentry"]["message"] == "this is a log with a dictionary %s"
    assert (
        event["logentry"]["formatted"]
        == "this is a log with a dictionary {'foo': 'bar'}"
    )
    assert event["logentry"]["params"] == {"foo": "bar"}


def test_logging_dictionary_args(sentry_init, capture_events):
    """Here we test items from a dictionary being interpolated into the log message."""
    sentry_init(integrations=[LoggingIntegration()], default_integrations=False)
    events = capture_events()

    logger.error(
        "the value of foo is %(foo)s, and the value of bar is %(bar)s",
        {"foo": "bar", "bar": "baz"},
    )

    (event,) = events
    assert (
        event["logentry"]["message"]
        == "the value of foo is %(foo)s, and the value of bar is %(bar)s"
    )
    assert (
        event["logentry"]["formatted"]
        == "the value of foo is bar, and the value of bar is baz"
    )
    assert event["logentry"]["params"] == {"foo": "bar", "bar": "baz"}


def test_sentry_logs_warning(sentry_init, capture_items):
    """
    The python logger module should create 'warn' sentry logs if the flag is on.
    """
    sentry_init(enable_logs=True)
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.warning("this is %s a template %s", "1", "2")

    get_client().flush()
    logs = [item.payload for item in items]
    attrs = logs[0]["attributes"]
    assert attrs["sentry.message.template"] == "this is %s a template %s"
    assert "code.file.path" in attrs
    assert "code.line.number" in attrs
    assert attrs["logger.name"] == "test-logger"
    assert attrs["sentry.environment"] == "production"
    assert attrs["sentry.message.parameter.0"] == "1"
    assert attrs["sentry.message.parameter.1"] == "2"
    assert attrs["sentry.origin"] == "auto.log.stdlib"
    assert logs[0]["attributes"]["sentry.severity_number"] == 13
    assert logs[0]["attributes"]["sentry.severity_text"] == "warn"


def test_sentry_logs_debug(sentry_init, capture_envelopes):
    """
    The python logger module should not create 'debug' sentry logs if the flag is on by default
    """
    sentry_init(enable_logs=True)
    envelopes = capture_envelopes()

    python_logger = logging.Logger("test-logger")
    python_logger.debug("this is %s a template %s", "1", "2")
    get_client().flush()

    assert len(envelopes) == 0


def test_no_log_infinite_loop(sentry_init, capture_envelopes):
    """
    If 'debug' mode is true, and you set a low log level in the logging integration, there should be no infinite loops.
    """
    sentry_init(
        enable_logs=True,
        integrations=[LoggingIntegration(sentry_logs_level=logging.DEBUG)],
        debug=True,
    )
    envelopes = capture_envelopes()

    python_logger = logging.Logger("test-logger")
    python_logger.debug("this is %s a template %s", "1", "2")
    get_client().flush()

    assert len(envelopes) == 1


def test_logging_errors(sentry_init, capture_envelopes, capture_items):
    """
    The python logger module should be able to log errors without erroring
    """
    sentry_init(enable_logs=True)
    envelopes = capture_envelopes()
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.error(Exception("test exc 1"))
    python_logger.error("error is %s", Exception("test exc 2"))
    get_client().flush()

    error_event_1 = envelopes[0].items[0].payload.json
    assert error_event_1["level"] == "error"
    error_event_2 = envelopes[1].items[0].payload.json
    assert error_event_2["level"] == "error"

    logs = [item.payload for item in items]
    assert logs[0]["attributes"]["sentry.severity_text"] == "error"
    assert "sentry.message.template" not in logs[0]["attributes"]
    assert "sentry.message.parameter.0" not in logs[0]["attributes"]
    assert "code.line.number" in logs[0]["attributes"]

    assert logs[1]["attributes"]["sentry.severity_text"] == "error"
    assert logs[1]["attributes"]["sentry.message.template"] == "error is %s"
    assert logs[1]["attributes"]["sentry.message.parameter.0"] in (
        "Exception('test exc 2')",
        "Exception('test exc 2',)",  # py3.6
    )
    assert "code.line.number" in logs[1]["attributes"]

    assert len(logs) == 2


def test_log_strips_project_root(sentry_init, capture_items):
    """
    The python logger should strip project roots from the log record path
    """
    sentry_init(
        enable_logs=True,
        project_root="/custom/test",
    )
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.handle(
        logging.LogRecord(
            name="test-logger",
            level=logging.WARN,
            pathname="/custom/test/blah/path.py",
            lineno=123,
            msg="This is a test log with a custom pathname",
            args=(),
            exc_info=None,
        )
    )
    get_client().flush()

    logs = [item.payload for item in items]
    assert len(logs) == 1
    attrs = logs[0]["attributes"]
    assert attrs["code.file.path"] == "blah/path.py"


def test_logger_with_all_attributes(sentry_init, capture_items):
    """
    The python logger should be able to log all attributes, including extra data.
    """
    sentry_init(enable_logs=True)
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.warning(
        "log #%d",
        1,
        extra={"foo": "bar", "numeric": 42, "more_complex": {"nested": "data"}},
    )
    get_client().flush()

    logs = [item.payload for item in items]

    assert "span_id" not in logs[0]

    attributes = logs[0]["attributes"]

    assert "process.pid" in attributes
    assert isinstance(attributes["process.pid"], int)
    del attributes["process.pid"]

    assert "sentry.release" in attributes
    assert isinstance(attributes["sentry.release"], str)
    del attributes["sentry.release"]

    assert "server.address" in attributes
    assert isinstance(attributes["server.address"], str)
    del attributes["server.address"]

    assert "thread.id" in attributes
    assert isinstance(attributes["thread.id"], int)
    del attributes["thread.id"]

    assert "code.file.path" in attributes
    assert isinstance(attributes["code.file.path"], str)
    del attributes["code.file.path"]

    assert "code.function.name" in attributes
    assert isinstance(attributes["code.function.name"], str)
    del attributes["code.function.name"]

    assert "code.line.number" in attributes
    assert isinstance(attributes["code.line.number"], int)
    del attributes["code.line.number"]

    assert "process.executable.name" in attributes
    assert isinstance(attributes["process.executable.name"], str)
    del attributes["process.executable.name"]

    assert "thread.name" in attributes
    assert isinstance(attributes["thread.name"], str)
    del attributes["thread.name"]

    assert attributes.pop("sentry.sdk.name").startswith("sentry.python")

    # Assert on the remaining non-dynamic attributes.
    assert attributes == {
        "foo": "bar",
        "numeric": 42,
        "more_complex": "{'nested': 'data'}",
        "logger.name": "test-logger",
        "sentry.origin": "auto.log.stdlib",
        "sentry.message.template": "log #%d",
        "process.runtime.name": mock.ANY,
        "process.runtime.version": mock.ANY,
        "sentry.message.parameter.0": 1,
        "sentry.environment": "production",
        "sentry.sdk.version": VERSION,
        "sentry.severity_number": 13,
        "sentry.severity_text": "warn",
    }


def test_sentry_logs_named_parameters(sentry_init, capture_items):
    """
    The python logger module should capture named parameters from dictionary arguments in Sentry logs.
    """
    sentry_init(enable_logs=True)
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.info(
        "%(source)s call completed, %(input_tk)i input tk, %(output_tk)i output tk (model %(model)s, cost $%(cost).4f)",
        {
            "source": "test_source",
            "input_tk": 100,
            "output_tk": 50,
            "model": "gpt-4",
            "cost": 0.0234,
        },
    )

    get_client().flush()
    logs = [item.payload for item in items]

    assert len(logs) == 1
    attrs = logs[0]["attributes"]

    # Check that the template is captured
    assert (
        attrs["sentry.message.template"]
        == "%(source)s call completed, %(input_tk)i input tk, %(output_tk)i output tk (model %(model)s, cost $%(cost).4f)"
    )

    # Check that dictionary arguments are captured as named parameters
    assert attrs["sentry.message.parameter.source"] == "test_source"
    assert attrs["sentry.message.parameter.input_tk"] == 100
    assert attrs["sentry.message.parameter.output_tk"] == 50
    assert attrs["sentry.message.parameter.model"] == "gpt-4"
    assert attrs["sentry.message.parameter.cost"] == 0.0234

    # Check other standard attributes
    assert attrs["logger.name"] == "test-logger"
    assert attrs["sentry.origin"] == "auto.log.stdlib"
    assert logs[0]["attributes"]["sentry.severity_number"] == 9  # info level
    assert logs[0]["attributes"]["sentry.severity_text"] == "info"


def test_sentry_logs_named_parameters_complex_values(sentry_init, capture_items):
    """
    The python logger module should handle complex values in named parameters using safe_repr.
    """
    sentry_init(enable_logs=True)
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    complex_object = {"nested": {"data": [1, 2, 3]}, "tuple": (4, 5, 6)}
    python_logger.warning(
        "Processing %(simple)s with %(complex)s data",
        {
            "simple": "simple_value",
            "complex": complex_object,
        },
    )

    get_client().flush()
    logs = [item.payload for item in items]

    assert len(logs) == 1
    attrs = logs[0]["attributes"]

    # Check that simple values are kept as-is
    assert attrs["sentry.message.parameter.simple"] == "simple_value"

    # Check that complex values are converted using safe_repr
    assert "sentry.message.parameter.complex" in attrs
    complex_param = attrs["sentry.message.parameter.complex"]
    assert isinstance(complex_param, str)
    assert "nested" in complex_param
    assert "data" in complex_param


def test_sentry_logs_no_parameters_no_template(sentry_init, capture_items):
    """
    There shouldn't be a template if there are no parameters.
    """
    sentry_init(enable_logs=True)
    items = capture_items("log")

    python_logger = logging.Logger("test-logger")
    python_logger.warning("Warning about something without any parameters.")

    get_client().flush()
    logs = [item.payload for item in items]

    assert len(logs) == 1

    attrs = logs[0]["attributes"]
    assert "sentry.message.template" not in attrs



import re
import sys
import threading
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk._queue import Queue
from sentry_sdk.integrations import Integration
from sentry_sdk.utils import (
    Components,
    Dsn,
    _get_installed_modules,
    datetime_from_isoformat,
    ensure_integration_enabled,
    env_to_bool,
    exc_info_from_error,
    format_timestamp,
    get_current_thread_meta,
    get_default_release,
    get_error_message,
    get_git_revision,
    get_lines_from_file,
    is_sentry_url,
    is_valid_sample_rate,
    logger,
    match_regex_list,
    package_version,
    parse_url,
    parse_version,
    safe_serialize,
    safe_str,
    sanitize_url,
    serialize_frame,
    to_string,
)


class TestIntegration(Integration):
    """
    Test integration for testing ensure_integration_enabled decorator.
    """

    identifier = "test"
    setup_once = mock.MagicMock()


try:
    import gevent
except ImportError:
    gevent = None


def _normalize_distribution_name(name: str) -> str:
    """Normalize distribution name according to PEP-0503.

    See:
    https://peps.python.org/pep-0503/#normalized-names
    for more details.
    """
    return re.sub(r"[-_.]+", "-", name).lower()


isoformat_inputs_and_datetime_outputs = (
    (
        "2021-01-01T00:00:00.000000Z",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),  # UTC time
    (
        "2021-01-01T00:00:00.000000",
        datetime(2021, 1, 1).astimezone(timezone.utc),
    ),  # No TZ -- assume local but convert to UTC
    (
        "2021-01-01T00:00:00Z",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),  # UTC - No milliseconds
    (
        "2021-01-01T00:00:00.000000+00:00",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),
    (
        "2021-01-01T00:00:00.000000-00:00",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),
    (
        "2021-01-01T00:00:00.000000+0000",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),
    (
        "2021-01-01T00:00:00.000000-0000",
        datetime(2021, 1, 1, tzinfo=timezone.utc),
    ),
    (
        "2020-12-31T00:00:00.000000+02:00",
        datetime(2020, 12, 31, tzinfo=timezone(timedelta(hours=2))),
    ),  # UTC+2 time
    (
        "2020-12-31T00:00:00.000000-0200",
        datetime(2020, 12, 31, tzinfo=timezone(timedelta(hours=-2))),
    ),  # UTC-2 time
    (
        "2020-12-31T00:00:00-0200",
        datetime(2020, 12, 31, tzinfo=timezone(timedelta(hours=-2))),
    ),  # UTC-2 time - no milliseconds
)


@pytest.mark.parametrize(
    ("input_str", "expected_output"),
    isoformat_inputs_and_datetime_outputs,
)
def test_datetime_from_isoformat(input_str, expected_output):
    assert datetime_from_isoformat(input_str) == expected_output, input_str


@pytest.mark.parametrize(
    ("input_str", "expected_output"),
    isoformat_inputs_and_datetime_outputs,
)
def test_datetime_from_isoformat_with_py_36_or_lower(input_str, expected_output):
    """
    `fromisoformat` was added in Python version 3.7
    """
    with mock.patch("sentry_sdk.utils.datetime") as datetime_mocked:
        datetime_mocked.fromisoformat.side_effect = AttributeError()
        datetime_mocked.strptime = datetime.strptime
        assert datetime_from_isoformat(input_str) == expected_output, input_str


@pytest.mark.parametrize(
    "env_var_value,strict,expected",
    [
        (None, True, None),
        (None, False, False),
        ("", True, None),
        ("", False, False),
        ("t", True, True),
        ("T", True, True),
        ("t", False, True),
        ("T", False, True),
        ("y", True, True),
        ("Y", True, True),
        ("y", False, True),
        ("Y", False, True),
        ("1", True, True),
        ("1", False, True),
        ("True", True, True),
        ("True", False, True),
        ("true", True, True),
        ("true", False, True),
        ("tRuE", True, True),
        ("tRuE", False, True),
        ("Yes", True, True),
        ("Yes", False, True),
        ("yes", True, True),
        ("yes", False, True),
        ("yEs", True, True),
        ("yEs", False, True),
        ("On", True, True),
        ("On", False, True),
        ("on", True, True),
        ("on", False, True),
        ("oN", True, True),
        ("oN", False, True),
        ("f", True, False),
        ("f", False, False),
        ("n", True, False),
        ("N", True, False),
        ("n", False, False),
        ("N", False, False),
        ("0", True, False),
        ("0", False, False),
        ("False", True, False),
        ("False", False, False),
        ("false", True, False),
        ("false", False, False),
        ("FaLsE", True, False),
        ("FaLsE", False, False),
        ("No", True, False),
        ("No", False, False),
        ("no", True, False),
        ("no", False, False),
        ("nO", True, False),
        ("nO", False, False),
        ("Off", True, False),
        ("Off", False, False),
        ("off", True, False),
        ("off", False, False),
        ("oFf", True, False),
        ("oFf", False, False),
        ("xxx", True, None),
        ("xxx", False, True),
    ],
)
def test_env_to_bool(env_var_value, strict, expected):
    assert env_to_bool(env_var_value, strict=strict) == expected, (
        f"Value: {env_var_value}, strict: {strict}"
    )


@pytest.mark.parametrize(
    ("url", "expected_result"),
    [
        ("http://localhost:8000", "http://localhost:8000"),
        ("http://example.com", "http://example.com"),
        ("https://example.com", "https://example.com"),
        (
            "example.com?token=abc&sessionid=123&save=true",
            "example.com?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
        (
            "http://example.com?token=abc&sessionid=123&save=true",
            "http://example.com?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
        (
            "https://example.com?token=abc&sessionid=123&save=true",
            "https://example.com?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
        (
            "http://localhost:8000/?token=abc&sessionid=123&save=true",
            "http://localhost:8000/?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
        (
            "ftp://username:password@ftp.example.com:9876/bla/blub#foo",
            "ftp://[Filtered]:[Filtered]@ftp.example.com:9876/bla/blub#foo",
        ),
        (
            "https://username:password@example.com/bla/blub?token=abc&sessionid=123&save=true#fragment",
            "https://[Filtered]:[Filtered]@example.com/bla/blub?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]#fragment",
        ),
        ("bla/blub/foo", "bla/blub/foo"),
        ("/bla/blub/foo/", "/bla/blub/foo/"),
        (
            "bla/blub/foo?token=abc&sessionid=123&save=true",
            "bla/blub/foo?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
        (
            "/bla/blub/foo/?token=abc&sessionid=123&save=true",
            "/bla/blub/foo/?token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
        ),
    ],
)
def test_sanitize_url(url, expected_result):
    assert sanitize_url(url) == expected_result


@pytest.mark.parametrize(
    ("url", "expected_result"),
    [
        (
            "http://localhost:8000",
            Components(
                scheme="http", netloc="localhost:8000", path="", query="", fragment=""
            ),
        ),
        (
            "http://example.com",
            Components(
                scheme="http", netloc="example.com", path="", query="", fragment=""
            ),
        ),
        (
            "https://example.com",
            Components(
                scheme="https", netloc="example.com", path="", query="", fragment=""
            ),
        ),
        (
            "example.com?token=abc&sessionid=123&save=true",
            Components(
                scheme="",
                netloc="",
                path="example.com",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
        (
            "http://example.com?token=abc&sessionid=123&save=true",
            Components(
                scheme="http",
                netloc="example.com",
                path="",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
        (
            "https://example.com?token=abc&sessionid=123&save=true",
            Components(
                scheme="https",
                netloc="example.com",
                path="",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
        (
            "http://localhost:8000/?token=abc&sessionid=123&save=true",
            Components(
                scheme="http",
                netloc="localhost:8000",
                path="/",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
        (
            "ftp://username:password@ftp.example.com:9876/bla/blub#foo",
            Components(
                scheme="ftp",
                netloc="[Filtered]:[Filtered]@ftp.example.com:9876",
                path="/bla/blub",
                query="",
                fragment="foo",
            ),
        ),
        (
            "https://username:password@example.com/bla/blub?token=abc&sessionid=123&save=true#fragment",
            Components(
                scheme="https",
                netloc="[Filtered]:[Filtered]@example.com",
                path="/bla/blub",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="fragment",
            ),
        ),
        (
            "bla/blub/foo",
            Components(
                scheme="", netloc="", path="bla/blub/foo", query="", fragment=""
            ),
        ),
        (
            "bla/blub/foo?token=abc&sessionid=123&save=true",
            Components(
                scheme="",
                netloc="",
                path="bla/blub/foo",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
        (
            "/bla/blub/foo/?token=abc&sessionid=123&save=true",
            Components(
                scheme="",
                netloc="",
                path="/bla/blub/foo/",
                query="token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
                fragment="",
            ),
        ),
    ],
)
def test_sanitize_url_and_split(url, expected_result):
    sanitized_url = sanitize_url(url, split=True)

    assert sanitized_url.scheme == expected_result.scheme
    assert sanitized_url.netloc == expected_result.netloc
    assert sanitized_url.query == expected_result.query
    assert sanitized_url.path == expected_result.path
    assert sanitized_url.fragment == expected_result.fragment


def test_sanitize_url_remove_authority_is_false():
    url = "https://usr:pwd@example.com"
    sanitized_url = sanitize_url(url, remove_authority=False)
    assert sanitized_url == url


@pytest.mark.parametrize(
    ("url", "sanitize", "expected_url", "expected_query", "expected_fragment"),
    [
        # Test with sanitize=True
        (
            "https://example.com",
            True,
            "https://example.com",
            "",
            "",
        ),
        (
            "example.com?token=abc&sessionid=123&save=true",
            True,
            "example.com",
            "token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
            "",
        ),
        (
            "https://example.com?token=abc&sessionid=123&save=true",
            True,
            "https://example.com",
            "token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
            "",
        ),
        (
            "https://username:password@example.com/bla/blub?token=abc&sessionid=123&save=true#fragment",
            True,
            "https://[Filtered]:[Filtered]@example.com/bla/blub",
            "token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
            "fragment",
        ),
        (
            "bla/blub/foo",
            True,
            "bla/blub/foo",
            "",
            "",
        ),
        (
            "/bla/blub/foo/#baz",
            True,
            "/bla/blub/foo/",
            "",
            "baz",
        ),
        (
            "bla/blub/foo?token=abc&sessionid=123&save=true",
            True,
            "bla/blub/foo",
            "token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
            "",
        ),
        (
            "/bla/blub/foo/?token=abc&sessionid=123&save=true",
            True,
            "/bla/blub/foo/",
            "token=[Filtered]&sessionid=[Filtered]&save=[Filtered]",
            "",
        ),
        # Test with sanitize=False
        (
            "https://example.com",
            False,
            "https://example.com",
            "",
            "",
        ),
        (
            "example.com?token=abc&sessionid=123&save=true",
            False,
            "example.com",
            "token=abc&sessionid=123&save=true",
            "",
        ),
        (
            "https://example.com?token=abc&sessionid=123&save=true",
            False,
            "https://example.com",
            "token=abc&sessionid=123&save=true",
            "",
        ),
        (
            "https://username:password@example.com/bla/blub?token=abc&sessionid=123&save=true#fragment",
            False,
            "https://[Filtered]:[Filtered]@example.com/bla/blub",
            "token=abc&sessionid=123&save=true",
            "fragment",
        ),
        (
            "bla/blub/foo",
            False,
            "bla/blub/foo",
            "",
            "",
        ),
        (
            "/bla/blub/foo/#baz",
            False,
            "/bla/blub/foo/",
            "",
            "baz",
        ),
        (
            "bla/blub/foo?token=abc&sessionid=123&save=true",
            False,
            "bla/blub/foo",
            "token=abc&sessionid=123&save=true",
            "",
        ),
        (
            "/bla/blub/foo/?token=abc&sessionid=123&save=true",
            False,
            "/bla/blub/foo/",
            "token=abc&sessionid=123&save=true",
            "",
        ),
    ],
)
def test_parse_url(url, sanitize, expected_url, expected_query, expected_fragment):
    assert parse_url(url, sanitize=sanitize).url == expected_url
    assert parse_url(url, sanitize=sanitize).fragment == expected_fragment
    assert parse_url(url, sanitize=sanitize).query == expected_query


@pytest.mark.parametrize(
    "rate",
    [0.0, 0.1231, 1.0, True, False],
)
def test_accepts_valid_sample_rate(rate):
    with mock.patch.object(logger, "warning", mock.Mock()):
        result = is_valid_sample_rate(rate, source="Testing")
        assert logger.warning.called is False
        assert result is True


@pytest.mark.parametrize(
    "rate",
    [
        "dogs are great",  # wrong type
        (0, 1),  # wrong type
        {"Maisey": "Charllie"},  # wrong type
        [True, True],  # wrong type
        {0.2012},  # wrong type
        float("NaN"),  # wrong type
        None,  # wrong type
        -1.121,  # wrong value
        1.231,  # wrong value
    ],
)
def test_warns_on_invalid_sample_rate(rate, StringContaining):  # noqa: N803
    with mock.patch.object(logger, "warning", mock.Mock()):
        result = is_valid_sample_rate(rate, source="Testing")
        logger.warning.assert_any_call(StringContaining("Given sample rate is invalid"))
        assert result is False


@pytest.mark.parametrize(
    "include_source_context",
    [True, False],
)
def test_include_source_context_when_serializing_frame(include_source_context):
    frame = sys._getframe()
    result = serialize_frame(frame, include_source_context=include_source_context)

    assert include_source_context ^ ("pre_context" in result) ^ True
    assert include_source_context ^ ("context_line" in result) ^ True
    assert include_source_context ^ ("post_context" in result) ^ True


@pytest.mark.parametrize(
    "item,regex_list,expected_result",
    [
        ["", [], False],
        [None, [], False],
        ["", None, False],
        [None, None, False],
        ["some-string", [], False],
        ["some-string", None, False],
        ["some-string", ["some-string"], True],
        ["some-string", ["some"], False],
        ["some-string", ["some$"], False],  # same as above
        ["some-string", ["some.*"], True],
        ["some-string", ["Some"], False],  # we do case sensitive matching
        ["some-string", [".*string$"], True],
    ],
)
def test_match_regex_list(item, regex_list, expected_result):
    assert match_regex_list(item, regex_list) == expected_result


@pytest.mark.parametrize(
    "version,expected_result",
    [
        ["3.5.15", (3, 5, 15)],
        ["2.0.9", (2, 0, 9)],
        ["2.0.0", (2, 0, 0)],
        ["0.6.0", (0, 6, 0)],
        ["2.0.0.post1", (2, 0, 0)],
        ["2.0.0rc3", (2, 0, 0)],
        ["2.0.0rc2", (2, 0, 0)],
        ["2.0.0rc1", (2, 0, 0)],
        ["2.0.0b4", (2, 0, 0)],
        ["2.0.0b3", (2, 0, 0)],
        ["2.0.0b2", (2, 0, 0)],
        ["2.0.0b1", (2, 0, 0)],
        ["0.6beta3", (0, 6)],
        ["0.6beta2", (0, 6)],
        ["0.6beta1", (0, 6)],
        ["0.4.2b", (0, 4, 2)],
        ["0.4.2a", (0, 4, 2)],
        ["0.0.1", (0, 0, 1)],
        ["0.0.0", (0, 0, 0)],
        ["1", (1,)],
        ["1.0", (1, 0)],
        ["1.0.0", (1, 0, 0)],
        [" 1.0.0 ", (1, 0, 0)],
        ["  1.0.0   ", (1, 0, 0)],
        ["x1.0.0", None],
        ["1.0.0x", None],
        ["x1.0.0x", None],
    ],
)
def test_parse_version(version, expected_result):
    assert parse_version(version) == expected_result


@pytest.fixture
def mock_client_with_dsn_netloc():
    """
    Returns a mocked Client with a DSN netloc of "abcd1234.ingest.sentry.io".
    """
    mock_client = mock.Mock(spec=sentry_sdk.Client)
    mock_client.transport = mock.Mock(spec=sentry_sdk.Transport)
    mock_client.transport.parsed_dsn = mock.Mock(spec=Dsn)

    mock_client.transport.parsed_dsn.netloc = "abcd1234.ingest.sentry.io"

    return mock_client


@pytest.mark.parametrize(
    ["test_url", "is_sentry_url_expected"],
    [
        ["https://asdf@abcd1234.ingest.sentry.io/123456789", True],
        ["https://asdf@abcd1234.ingest.notsentry.io/123456789", False],
    ],
)
def test_is_sentry_url_true(
    test_url, is_sentry_url_expected, mock_client_with_dsn_netloc
):
    ret_val = is_sentry_url(mock_client_with_dsn_netloc, test_url)

    assert ret_val == is_sentry_url_expected


def test_is_sentry_url_no_client():
    test_url = "https://asdf@abcd1234.ingest.sentry.io/123456789"

    ret_val = is_sentry_url(None, test_url)

    assert not ret_val


@pytest.mark.parametrize(
    "error,expected_result",
    [
        ["", lambda x: safe_str(x)],
        ["some-string", lambda _: "some-string"],
    ],
)
def test_get_error_message(error, expected_result):
    with pytest.raises(BaseException) as exc_value:
        exc_value.message = error
        raise Exception
    assert get_error_message(exc_value) == expected_result(exc_value)

    with pytest.raises(BaseException) as exc_value:
        exc_value.detail = error
        raise Exception
    assert get_error_message(exc_value) == expected_result(exc_value)


def test_safe_str_fails():
    class ExplodingStr:
        def __str__(self):
            raise Exception

    obj = ExplodingStr()
    result = safe_str(obj)

    assert result == repr(obj)


def test_installed_modules_caching():
    mock_generate_installed_modules = mock.Mock()
    mock_generate_installed_modules.return_value = {"package": "1.0.0"}
    with mock.patch("sentry_sdk.utils._installed_modules", None):
        with mock.patch(
            "sentry_sdk.utils._generate_installed_modules",
            mock_generate_installed_modules,
        ):
            _get_installed_modules()
            assert mock_generate_installed_modules.called
            mock_generate_installed_modules.reset_mock()

            _get_installed_modules()
            mock_generate_installed_modules.assert_not_called()


def test_devnull_inaccessible():
    with mock.patch("sentry_sdk.utils.open", side_effect=OSError("oh no")):
        revision = get_git_revision()

    assert revision is None


def test_devnull_not_found():
    with mock.patch("sentry_sdk.utils.open", side_effect=FileNotFoundError("oh no")):
        revision = get_git_revision()

    assert revision is None


def test_default_release():
    release = get_default_release()
    assert release is not None


def test_default_release_empty_string():
    with mock.patch("sentry_sdk.utils.get_git_revision", return_value=""):
        release = get_default_release()

    assert release is None


def test_get_default_release_sentry_release_env(monkeypatch):
    monkeypatch.setenv("SENTRY_RELEASE", "sentry-env-release")
    assert get_default_release() == "sentry-env-release"


def test_get_default_release_other_release_env(monkeypatch):
    monkeypatch.setenv("SOURCE_VERSION", "other-env-release")

    with mock.patch("sentry_sdk.utils.get_git_revision", return_value=""):
        release = get_default_release()

    assert release == "other-env-release"


def test_get_default_release_heroku_build_commit(monkeypatch):
    monkeypatch.setenv("HEROKU_BUILD_COMMIT", "heroku-build-commit-sha")

    with mock.patch("sentry_sdk.utils.get_git_revision", return_value=""):
        release = get_default_release()

    assert release == "heroku-build-commit-sha"


def test_get_default_release_heroku_slug_commit_fallback(monkeypatch):
    # Although deprecated by Heroku, HEROKU_SLUG_COMMIT should still be used if HEROKU_BUILD_COMMIT is not set
    monkeypatch.setenv("HEROKU_SLUG_COMMIT", "heroku-slug-commit-sha")

    with mock.patch("sentry_sdk.utils.get_git_revision", return_value=""):
        release = get_default_release()

    assert release == "heroku-slug-commit-sha"


def test_get_default_release_heroku_build_commit_takes_priority(monkeypatch):
    # HEROKU_BUILD_COMMIT should take priority over HEROKU_SLUG_COMMIT since it's the non-deprecated variable
    monkeypatch.setenv("HEROKU_BUILD_COMMIT", "heroku-build-commit-sha")
    monkeypatch.setenv("HEROKU_SLUG_COMMIT", "heroku-slug-commit-sha")

    with mock.patch("sentry_sdk.utils.get_git_revision", return_value=""):
        release = get_default_release()

    assert release == "heroku-build-commit-sha"


def test_ensure_integration_enabled_integration_enabled(sentry_init):
    def original_function():
        return "original"

    def function_to_patch():
        return "patched"

    sentry_init(integrations=[TestIntegration()])

    # Test the decorator by applying to function_to_patch
    patched_function = ensure_integration_enabled(TestIntegration, original_function)(
        function_to_patch
    )

    assert patched_function() == "patched"
    assert patched_function.__name__ == "original_function"


def test_ensure_integration_enabled_integration_disabled(sentry_init):
    def original_function():
        return "original"

    def function_to_patch():
        return "patched"

    sentry_init(integrations=[])  # TestIntegration is disabled

    # Test the decorator by applying to function_to_patch
    patched_function = ensure_integration_enabled(TestIntegration, original_function)(
        function_to_patch
    )

    assert patched_function() == "original"
    assert patched_function.__name__ == "original_function"


def test_ensure_integration_enabled_no_original_function_enabled(sentry_init):
    shared_variable = "original"

    def function_to_patch():
        nonlocal shared_variable
        shared_variable = "patched"

    sentry_init(integrations=[TestIntegration])

    # Test the decorator by applying to function_to_patch
    patched_function = ensure_integration_enabled(TestIntegration)(function_to_patch)
    patched_function()

    assert shared_variable == "patched"
    assert patched_function.__name__ == "function_to_patch"


def test_ensure_integration_enabled_no_original_function_disabled(sentry_init):
    shared_variable = "original"

    def function_to_patch():
        nonlocal shared_variable
        shared_variable = "patched"

    sentry_init(integrations=[])

    # Test the decorator by applying to function_to_patch
    patched_function = ensure_integration_enabled(TestIntegration)(function_to_patch)
    patched_function()

    assert shared_variable == "original"
    assert patched_function.__name__ == "function_to_patch"


@pytest.mark.parametrize(
    "delta,expected_milliseconds",
    [
        [timedelta(milliseconds=132), 132.0],
        [timedelta(hours=1, milliseconds=132), float(60 * 60 * 1000 + 132)],
        [timedelta(days=10), float(10 * 24 * 60 * 60 * 1000)],
        [timedelta(microseconds=100), 0.1],
    ],
)
def test_duration_in_milliseconds(delta, expected_milliseconds):
    assert delta / timedelta(milliseconds=1) == expected_milliseconds


def test_get_current_thread_meta_explicit_thread():
    results = Queue(maxsize=1)

    def target1():
        pass

    def target2():
        results.put(get_current_thread_meta(thread1))

    thread1 = threading.Thread(target=target1)
    thread1.start()

    thread2 = threading.Thread(target=target2)
    thread2.start()

    thread2.join()
    thread1.join()

    assert (thread1.ident, thread1.name) == results.get(timeout=1)


def test_get_current_thread_meta_bad_explicit_thread():
    thread = "fake thread"

    main_thread = threading.main_thread()

    assert (main_thread.ident, main_thread.name) == get_current_thread_meta(thread)


@pytest.mark.skipif(gevent is None, reason="gevent not enabled")
def test_get_current_thread_meta_gevent_in_thread():
    results = Queue(maxsize=1)

    def target():
        with mock.patch("sentry_sdk.utils.is_gevent", side_effect=[True]):
            job = gevent.spawn(get_current_thread_meta)
            job.join()
            results.put(job.value)

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()
    assert (thread.ident, None) == results.get(timeout=1)


@pytest.mark.skipif(gevent is None, reason="gevent not enabled")
def test_get_current_thread_meta_gevent_in_thread_failed_to_get_hub():
    results = Queue(maxsize=1)

    def target():
        with mock.patch("sentry_sdk.utils.is_gevent", side_effect=[True]):
            with mock.patch(
                "sentry_sdk.utils.get_gevent_hub", side_effect=["fake gevent hub"]
            ):
                job = gevent.spawn(get_current_thread_meta)
                job.join()
                results.put(job.value)

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()
    assert (thread.ident, thread.name) == results.get(timeout=1)


def test_get_current_thread_meta_running_thread():
    results = Queue(maxsize=1)

    def target():
        results.put(get_current_thread_meta())

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()
    assert (thread.ident, thread.name) == results.get(timeout=1)


def test_get_current_thread_meta_bad_running_thread():
    results = Queue(maxsize=1)

    def target():
        with mock.patch("threading.current_thread", side_effect=["fake thread"]):
            results.put(get_current_thread_meta())

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()

    main_thread = threading.main_thread()
    assert (main_thread.ident, main_thread.name) == results.get(timeout=1)


def test_get_current_thread_meta_main_thread():
    results = Queue(maxsize=1)

    def target():
        # mock that somehow the current thread doesn't exist
        with mock.patch("threading.current_thread", side_effect=[None]):
            results.put(get_current_thread_meta())

    main_thread = threading.main_thread()

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()
    assert (main_thread.ident, main_thread.name) == results.get(timeout=1)


def test_get_current_thread_meta_failed_to_get_main_thread():
    results = Queue(maxsize=1)

    def target():
        with mock.patch("threading.current_thread", return_value="fake thread"):
            results.put(get_current_thread_meta())

    main_thread = threading.main_thread()

    thread = threading.Thread(target=target)
    thread.start()
    thread.join()
    assert (main_thread.ident, main_thread.name) == results.get(timeout=1)


@pytest.mark.parametrize(
    ("datetime_object", "expected_output"),
    (
        (
            datetime(2021, 1, 1, tzinfo=timezone.utc),
            "2021-01-01T00:00:00.000000Z",
        ),  # UTC time
        (
            datetime(2021, 1, 1, tzinfo=timezone(timedelta(hours=2))),
            "2020-12-31T22:00:00.000000Z",
        ),  # UTC+2 time
        (
            datetime(2021, 1, 1, tzinfo=timezone(timedelta(hours=-7))),
            "2021-01-01T07:00:00.000000Z",
        ),  # UTC-7 time
        (
            datetime(2021, 2, 3, 4, 56, 7, 890123, tzinfo=timezone.utc),
            "2021-02-03T04:56:07.890123Z",
        ),  # UTC time all non-zero fields
    ),
)
def test_format_timestamp(datetime_object, expected_output):
    formatted = format_timestamp(datetime_object)

    assert formatted == expected_output


def test_format_timestamp_naive():
    datetime_object = datetime(2021, 1, 1)
    timestamp_regex = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{6}Z"

    # Ensure that some timestamp is returned, without error. We currently treat these as local time, but this is an
    # implementation detail which we should not assert here.
    assert re.fullmatch(timestamp_regex, format_timestamp(datetime_object))


def test_qualname_from_function_inner_function():
    def test_function(): ...

    assert (
        sentry_sdk.utils.qualname_from_function(test_function)
        == "tests.test_utils.test_qualname_from_function_inner_function.<locals>.test_function"
    )


def test_qualname_from_function_none_name():
    def test_function(): ...

    test_function.__module__ = None

    assert (
        sentry_sdk.utils.qualname_from_function(test_function)
        == "test_qualname_from_function_none_name.<locals>.test_function"
    )


def test_to_string_unicode_decode_error():
    class BadStr:
        def __str__(self):
            raise UnicodeDecodeError("utf-8", b"", 0, 1, "reason")

    obj = BadStr()
    result = to_string(obj)
    assert result == repr(obj)[1:-1]


def test_exc_info_from_error_dont_get_an_exc():
    class NotAnException:
        pass

    with pytest.raises(ValueError) as exc:
        exc_info_from_error(NotAnException())

    assert "Expected Exception object to report, got <class" in str(exc.value)


def test_get_lines_from_file_handle_linecache_errors():
    expected_result = ([], None, [])

    class Loader:
        @staticmethod
        def get_source(module):
            raise IOError("something went wrong")

    result = get_lines_from_file("filename", 10, loader=Loader())
    assert result == expected_result

    with mock.patch(
        "sentry_sdk.utils.linecache.getlines",
        side_effect=OSError("something went wrong"),
    ):
        result = get_lines_from_file("filename", 10)
        assert result == expected_result

    lines = ["line1", "line2", "line3"]

    def fake_getlines(filename):
        return lines

    with mock.patch("sentry_sdk.utils.linecache.getlines", fake_getlines):
        result = get_lines_from_file("filename", 10)
        assert result == expected_result


def test_safe_serialize_plain_string():
    assert safe_serialize("already a string") == "already a string"


def test_safe_serialize_json_string():
    assert safe_serialize('{"key": "value"}') == '{"key": "value"}'


def test_safe_serialize_dict():
    assert safe_serialize({"key": "value"}) == '{"key": "value"}'


def test_safe_serialize_callable():
    def my_func():
        pass

    result = safe_serialize(my_func)
    assert result.startswith("<function")
    assert '"' not in result[:1]  # no wrapping quotes from json.dumps


def test_safe_serialize_object():
    class MyClass:
        def __init__(self):
            self.x = 1

    result = safe_serialize(MyClass())
    assert result.startswith("<MyClass")
    assert '"' not in result[:1]  # no wrapping quotes from json.dumps


def test_package_version_is_none():
    assert package_version("non_existent_package") is None



import json
import unittest.mock

import pytest
import trytond
from trytond.exceptions import LoginException
from trytond.exceptions import TrytonException as TrytondBaseException
from trytond.exceptions import UserError as TrytondUserError
from trytond.exceptions import UserWarning as TrytondUserWarning
from trytond.wsgi import app as trytond_app
from werkzeug.test import Client

from sentry_sdk.integrations.trytond import TrytondWSGIIntegration
from tests.conftest import unpack_werkzeug_response


@pytest.fixture(scope="function")
def app(sentry_init):
    yield trytond_app


@pytest.fixture
def get_client(app):
    def inner():
        return Client(app)

    return inner


@pytest.mark.parametrize(
    "exception", [Exception("foo"), type("FooException", (Exception,), {})("bar")]
)
def test_exceptions_captured(
    sentry_init, app, capture_exceptions, get_client, exception
):
    sentry_init(integrations=[TrytondWSGIIntegration()])
    exceptions = capture_exceptions()

    unittest.mock.sentinel.exception = exception

    @app.route("/exception")
    def _(request):
        raise unittest.mock.sentinel.exception

    client = get_client()
    _ = client.get("/exception")

    (e,) = exceptions
    assert e is exception


@pytest.mark.parametrize(
    "exception",
    [
        TrytondUserError("title"),
        TrytondUserWarning("title", "details"),
        LoginException("title", "details"),
    ],
)
def test_trytonderrors_not_captured(
    sentry_init, app, capture_exceptions, get_client, exception
):
    sentry_init(integrations=[TrytondWSGIIntegration()])
    exceptions = capture_exceptions()

    unittest.mock.sentinel.exception = exception

    @app.route("/usererror")
    def _(request):
        raise unittest.mock.sentinel.exception

    client = get_client()
    _ = client.get("/usererror")

    assert not exceptions


@pytest.mark.skipif(
    trytond.__version__.split(".") < ["5", "4"], reason="At least Trytond-5.4 required"
)
def test_rpc_error_page(sentry_init, app, get_client):
    """Test that, after initializing the Trytond-SentrySDK integration
    a custom error handler can be registered to the Trytond WSGI app so as to
    inform the event identifiers to the Tryton RPC client"""

    sentry_init(integrations=[TrytondWSGIIntegration()])

    @app.route("/rpcerror", methods=["POST"])
    def _(request):
        raise Exception("foo")

    @app.error_handler
    def _(app, request, e):
        if isinstance(e, TrytondBaseException):
            return
        else:
            data = TrytondUserError("Sentry error.", str(e))
            return app.make_response(request, data)

    client = get_client()

    # This would look like a natural Tryton RPC call
    _data = dict(
        id=42,  # request sequence
        method="class.method",  # rpc call
        params=[
            [1234],  # ids
            ["bar", "baz"],  # values
            dict(  # context
                client="12345678-9abc-def0-1234-56789abc",
                groups=[1],
                language="ca",
                language_direction="ltr",
            ),
        ],
    )
    response = client.post(
        "/rpcerror", content_type="application/json", data=json.dumps(_data)
    )

    (content, status, headers) = unpack_werkzeug_response(response)
    data = json.loads(content)
    assert status == "200 OK"
    assert headers.get("Content-Type") == "application/json"
    assert data == dict(id=42, error=["UserError", ["Sentry error.", "foo", None]])


def test_span_origin(sentry_init, app, capture_events, get_client):
    sentry_init(
        integrations=[TrytondWSGIIntegration()],
        traces_sample_rate=1.0,
    )
    events = capture_events()

    @app.route("/something")
    def _(request):
        return "ok"

    client = get_client()
    client.get("/something")

    (event,) = events

    assert event["contexts"]["trace"]["origin"] == "auto.http.trytond_wsgi"



import functools
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import (
    _DEFAULT_FAILED_REQUEST_STATUS_CODES,
    DidNotEnable,
    Integration,
    _check_minimum_version,
)
from sentry_sdk.integrations._wsgi_common import RequestExtractor
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware
from sentry_sdk.traces import SOURCE_FOR_STYLE as SEGMENT_SOURCE_FOR_STYLE
from sentry_sdk.tracing import SOURCE_FOR_STYLE as TRANSACTION_SOURCE_FOR_STYLE
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    ensure_integration_enabled,
    event_from_exception,
    parse_version,
    transaction_from_function,
)

if TYPE_CHECKING:
    from collections.abc import Set
    from typing import Any, Callable, Dict, Optional

    from bottle import FileUpload, FormsDict, LocalRequest  # type: ignore

    from sentry_sdk._types import Event, EventProcessor
    from sentry_sdk.integrations.wsgi import _ScopedResponse

try:
    from bottle import (
        Bottle,
        HTTPResponse,
        Route,
    )
    from bottle import (
        __version__ as BOTTLE_VERSION,
    )
    from bottle import (
        request as bottle_request,
    )
except ImportError:
    raise DidNotEnable("Bottle not installed")


TRANSACTION_STYLE_VALUES = ("endpoint", "url")


class BottleIntegration(Integration):
    identifier = "bottle"
    origin = f"auto.http.{identifier}"

    transaction_style = ""

    def __init__(
        self,
        transaction_style: str = "endpoint",
        *,
        failed_request_status_codes: "Set[int]" = _DEFAULT_FAILED_REQUEST_STATUS_CODES,
    ) -> None:
        if transaction_style not in TRANSACTION_STYLE_VALUES:
            raise ValueError(
                "Invalid value for transaction_style: %s (must be in %s)"
                % (transaction_style, TRANSACTION_STYLE_VALUES)
            )
        self.transaction_style = transaction_style
        self.failed_request_status_codes = failed_request_status_codes

    @staticmethod
    def setup_once() -> None:
        version = parse_version(BOTTLE_VERSION)
        _check_minimum_version(BottleIntegration, version)

        old_app = Bottle.__call__

        @ensure_integration_enabled(BottleIntegration, old_app)
        def sentry_patched_wsgi_app(
            self: "Any", environ: "Dict[str, str]", start_response: "Callable[..., Any]"
        ) -> "_ScopedResponse":
            middleware = SentryWsgiMiddleware(
                lambda *a, **kw: old_app(self, *a, **kw),
                span_origin=BottleIntegration.origin,
            )

            return middleware(environ, start_response)

        Bottle.__call__ = sentry_patched_wsgi_app

        old_handle = Bottle._handle

        @functools.wraps(old_handle)
        def _patched_handle(self: "Bottle", environ: "Dict[str, Any]") -> "Any":
            integration = sentry_sdk.get_client().get_integration(BottleIntegration)
            if integration is None:
                return old_handle(self, environ)

            scope = sentry_sdk.get_isolation_scope()
            scope._name = "bottle"
            scope.add_event_processor(
                _make_request_event_processor(self, bottle_request, integration)
            )
            res = old_handle(self, environ)

            if has_span_streaming_enabled(sentry_sdk.get_client().options):
                _set_segment_name_and_source(
                    transaction_style=integration.transaction_style
                )

            return res

        Bottle._handle = _patched_handle

        old_make_callback = Route._make_callback

        @functools.wraps(old_make_callback)
        def patched_make_callback(
            self: "Route", *args: object, **kwargs: object
        ) -> "Any":
            prepared_callback = old_make_callback(self, *args, **kwargs)

            integration = sentry_sdk.get_client().get_integration(BottleIntegration)
            if integration is None:
                return prepared_callback

            def wrapped_callback(*args: object, **kwargs: object) -> "Any":
                try:
                    res = prepared_callback(*args, **kwargs)
                except Exception as exception:
                    _capture_exception(exception, handled=False)
                    raise exception

                if (
                    isinstance(res, HTTPResponse)
                    and res.status_code in integration.failed_request_status_codes
                ):
                    _capture_exception(res, handled=True)

                return res

            return wrapped_callback

        Route._make_callback = patched_make_callback


class BottleRequestExtractor(RequestExtractor):
    def env(self) -> "Dict[str, str]":
        return self.request.environ

    def cookies(self) -> "Dict[str, str]":
        return self.request.cookies

    def raw_data(self) -> bytes:
        return self.request.body.read()

    def form(self) -> "FormsDict":
        if self.is_json():
            return None
        return self.request.forms.decode()

    def files(self) -> "Optional[Dict[str, str]]":
        if self.is_json():
            return None

        return self.request.files

    def size_of_file(self, file: "FileUpload") -> int:
        return file.content_length


def _set_segment_name_and_source(transaction_style: str) -> None:
    try:
        if transaction_style == "url":
            name = bottle_request.route.rule or "bottle request"
        else:
            name = (
                bottle_request.route.name
                or transaction_from_function(bottle_request.route.callback)
                or "bottle request"
            )

        sentry_sdk.get_current_scope().set_transaction_name(
            name,
            source=SEGMENT_SOURCE_FOR_STYLE[transaction_style],
        )
    except RuntimeError:
        pass


def _set_transaction_name_and_source(
    event: "Event", transaction_style: str, request: "Any"
) -> None:
    name = ""

    if transaction_style == "url":
        try:
            name = request.route.rule or ""
        except RuntimeError:
            pass

    elif transaction_style == "endpoint":
        try:
            name = (
                request.route.name
                or transaction_from_function(request.route.callback)
                or ""
            )
        except RuntimeError:
            pass

    event["transaction"] = name
    event["transaction_info"] = {
        "source": TRANSACTION_SOURCE_FOR_STYLE[transaction_style]
    }


def _make_request_event_processor(
    app: "Bottle", request: "LocalRequest", integration: "BottleIntegration"
) -> "EventProcessor":
    def event_processor(event: "Event", hint: "dict[str, Any]") -> "Event":
        _set_transaction_name_and_source(event, integration.transaction_style, request)

        with capture_internal_exceptions():
            BottleRequestExtractor(request).extract_into_event(event)

        return event

    return event_processor


def _capture_exception(exception: BaseException, handled: bool) -> None:
    event, hint = event_from_exception(
        exception,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "bottle", "handled": handled},
    )
    sentry_sdk.capture_event(event, hint=hint)



import asyncio
import inspect
import sys
from unittest.mock import MagicMock, Mock, patch

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock

import pytest

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations.asyncio import (
    AsyncioIntegration,
    enable_asyncio_integration,
    patch_asyncio,
)

try:
    from contextvars import Context, ContextVar
except ImportError:
    pass  # All tests will be skipped with incompatible versions


minimum_python_38 = pytest.mark.skipif(
    sys.version_info < (3, 8), reason="Asyncio tests need Python >= 3.8"
)


minimum_python_39 = pytest.mark.skipif(
    sys.version_info < (3, 9), reason="Test requires Python >= 3.9"
)


minimum_python_311 = pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="Asyncio task context parameter was introduced in Python 3.11",
)


async def foo():
    await asyncio.sleep(0.01)


async def bar():
    await asyncio.sleep(0.01)


async def boom():
    1 / 0


def get_sentry_task_factory(mock_get_running_loop):
    """
    Patches (mocked) asyncio and gets the sentry_task_factory.
    """
    mock_loop = mock_get_running_loop.return_value
    patch_asyncio()
    patched_factory = mock_loop.set_task_factory.call_args[0][0]

    return patched_factory


@minimum_python_38
@pytest.mark.asyncio(loop_scope="module")
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_create_task(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[
            AsyncioIntegration(),
        ],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(
            name="not so important", attributes={"sentry.op": "root"}
        ):
            foo_task = asyncio.create_task(foo())
            bar_task = asyncio.create_task(bar())

            if hasattr(foo_task.get_coro(), "__name__"):
                assert foo_task.get_coro().__name__ == "foo"
            if hasattr(bar_task.get_coro(), "__name__"):
                assert bar_task.get_coro().__name__ == "bar"

            tasks = [foo_task, bar_task]

            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction_for_create_task"):
            with sentry_sdk.start_span(op="root", name="not so important"):
                foo_task = asyncio.create_task(foo())
                bar_task = asyncio.create_task(bar())

                if hasattr(foo_task.get_coro(), "__name__"):
                    assert foo_task.get_coro().__name__ == "foo"
                if hasattr(bar_task.get_coro(), "__name__"):
                    assert bar_task.get_coro().__name__ == "bar"

                tasks = [foo_task, bar_task]

                await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    sentry_sdk.flush()

    if span_streaming:
        segment = items.pop().payload

        assert segment["is_segment"] is True
        assert segment["name"] == "not so important"
        assert segment["attributes"]["sentry.op"] == "root"

        spans = [item.payload for item in items]
        assert len(spans) == 2

        assert spans[0]["attributes"]["sentry.op"] == OP.FUNCTION
        assert spans[0]["name"] == "foo"
        assert spans[0]["parent_span_id"] == segment["span_id"]

        assert spans[1]["attributes"]["sentry.op"] == OP.FUNCTION
        assert spans[1]["name"] == "bar"
        assert spans[1]["parent_span_id"] == segment["span_id"]
    else:
        (transaction_event,) = events

        assert transaction_event["spans"][0]["op"] == "root"
        assert transaction_event["spans"][0]["description"] == "not so important"

        assert transaction_event["spans"][1]["op"] == OP.FUNCTION
        assert transaction_event["spans"][1]["description"] == "foo"
        assert (
            transaction_event["spans"][1]["parent_span_id"]
            == transaction_event["spans"][0]["span_id"]
        )

        assert transaction_event["spans"][2]["op"] == OP.FUNCTION
        assert transaction_event["spans"][2]["description"] == "bar"
        assert (
            transaction_event["spans"][2]["parent_span_id"]
            == transaction_event["spans"][0]["span_id"]
        )


@minimum_python_38
@pytest.mark.asyncio(loop_scope="module")
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_gather(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[
            AsyncioIntegration(),
        ],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(
            name="not so important", attributes={"sentry.op": "root"}
        ):
            await asyncio.gather(foo(), bar(), return_exceptions=True)

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction_for_gather"):
            with sentry_sdk.start_span(op="root", name="not so important"):
                await asyncio.gather(foo(), bar(), return_exceptions=True)

    sentry_sdk.flush()

    if span_streaming:
        segment = items.pop().payload
        assert segment["is_segment"] is True
        assert segment["name"] == "not so important"
        assert segment["attributes"]["sentry.op"] == "root"

        spans = [item.payload for item in items]
        assert len(spans) == 2

        assert spans[0]["attributes"]["sentry.op"] == OP.FUNCTION
        assert spans[0]["name"] == "foo"
        assert spans[0]["parent_span_id"] == segment["span_id"]

        assert spans[1]["attributes"]["sentry.op"] == OP.FUNCTION
        assert spans[1]["name"] == "bar"
        assert spans[1]["parent_span_id"] == segment["span_id"]
    else:
        (transaction_event,) = events

        assert transaction_event["spans"][0]["op"] == "root"
        assert transaction_event["spans"][0]["description"] == "not so important"

        assert transaction_event["spans"][1]["op"] == OP.FUNCTION
        assert transaction_event["spans"][1]["description"] == "foo"
        assert (
            transaction_event["spans"][1]["parent_span_id"]
            == transaction_event["spans"][0]["span_id"]
        )

        assert transaction_event["spans"][2]["op"] == OP.FUNCTION
        assert transaction_event["spans"][2]["description"] == "bar"
        assert (
            transaction_event["spans"][2]["parent_span_id"]
            == transaction_event["spans"][0]["span_id"]
        )


@minimum_python_38
@pytest.mark.asyncio(loop_scope="module")
async def test_exception(
    sentry_init,
    capture_events,
):
    sentry_init(
        traces_sample_rate=1.0,
        send_default_pii=True,
        integrations=[
            AsyncioIntegration(),
        ],
    )

    events = capture_events()

    with sentry_sdk.start_transaction(name="test_exception"):
        with sentry_sdk.start_span(op="root", name="not so important"):
            tasks = [asyncio.create_task(boom()), asyncio.create_task(bar())]
            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    sentry_sdk.flush()

    (error_event, _) = events

    assert error_event["transaction"] == "test_exception"
    assert error_event["contexts"]["trace"]["op"] == "function"
    assert error_event["exception"]["values"][0]["type"] == "ZeroDivisionError"
    assert error_event["exception"]["values"][0]["value"] == "division by zero"
    assert error_event["exception"]["values"][0]["mechanism"]["handled"] is False
    assert error_event["exception"]["values"][0]["mechanism"]["type"] == "asyncio"


@minimum_python_38
@pytest.mark.asyncio(loop_scope="module")
async def test_task_result(sentry_init):
    sentry_init(
        integrations=[
            AsyncioIntegration(),
        ],
    )

    async def add(a, b):
        return a + b

    result = await asyncio.create_task(add(1, 2))
    assert result == 3, result


@minimum_python_311
@pytest.mark.asyncio(loop_scope="module")
async def test_task_with_context(sentry_init):
    """
    Integration test to ensure working context parameter in Python 3.11+
    """
    sentry_init(
        integrations=[
            AsyncioIntegration(),
        ],
    )

    var = ContextVar("var")
    var.set("original value")

    async def change_value():
        var.set("changed value")

    async def retrieve_value():
        return var.get()

    # Create a context and run both tasks within the context
    ctx = Context()
    async with asyncio.TaskGroup() as tg:
        tg.create_task(change_value(), context=ctx)
        retrieve_task = tg.create_task(retrieve_value(), context=ctx)

    assert retrieve_task.result() == "changed value"


@minimum_python_38
@patch("asyncio.get_running_loop")
def test_patch_asyncio(mock_get_running_loop):
    """
    Test that the patch_asyncio function will patch the task factory.
    """
    mock_loop = mock_get_running_loop.return_value
    mock_loop.get_task_factory.return_value._is_sentry_task_factory = False

    patch_asyncio()

    assert mock_loop.set_task_factory.called

    set_task_factory_args, _ = mock_loop.set_task_factory.call_args
    assert len(set_task_factory_args) == 1

    sentry_task_factory, *_ = set_task_factory_args
    assert callable(sentry_task_factory)


@minimum_python_38
@patch("asyncio.get_running_loop")
@patch("sentry_sdk.integrations.asyncio.Task")
def test_sentry_task_factory_no_factory(MockTask, mock_get_running_loop):  # noqa: N803
    mock_loop = mock_get_running_loop.return_value
    mock_coro = MagicMock()

    # Set the original task factory to None
    mock_loop.get_task_factory.return_value = None

    # Retieve sentry task factory (since it is an inner function within patch_asyncio)
    sentry_task_factory = get_sentry_task_factory(mock_get_running_loop)

    # The call we are testing
    ret_val = sentry_task_factory(mock_loop, mock_coro)

    assert MockTask.called
    assert ret_val == MockTask.return_value

    task_args, task_kwargs = MockTask.call_args
    assert len(task_args) == 1

    coro_param, *_ = task_args
    assert inspect.iscoroutine(coro_param)

    assert "loop" in task_kwargs
    assert task_kwargs["loop"] == mock_loop


@minimum_python_38
@patch("asyncio.get_running_loop")
def test_sentry_task_factory_with_factory(mock_get_running_loop):
    mock_loop = mock_get_running_loop.return_value
    mock_coro = MagicMock()

    # The original task factory will be mocked out here, let's retrieve the value for later
    orig_task_factory = mock_loop.get_task_factory.return_value
    orig_task_factory._is_sentry_task_factory = False

    # Retieve sentry task factory (since it is an inner function within patch_asyncio)
    sentry_task_factory = get_sentry_task_factory(mock_get_running_loop)

    # The call we are testing
    ret_val = sentry_task_factory(mock_loop, mock_coro)

    assert orig_task_factory.called
    assert ret_val == orig_task_factory.return_value

    task_factory_args, _ = orig_task_factory.call_args
    assert len(task_factory_args) == 2

    loop_arg, coro_arg = task_factory_args
    assert loop_arg == mock_loop
    assert inspect.iscoroutine(coro_arg)


@minimum_python_311
@patch("asyncio.get_running_loop")
@patch("sentry_sdk.integrations.asyncio.Task")
def test_sentry_task_factory_context_no_factory(
    MockTask,
    mock_get_running_loop,  # noqa: N803
):
    mock_loop = mock_get_running_loop.return_value
    mock_coro = MagicMock()
    mock_context = MagicMock()

    # Set the original task factory to None
    mock_loop.get_task_factory.return_value = None

    # Retieve sentry task factory (since it is an inner function within patch_asyncio)
    sentry_task_factory = get_sentry_task_factory(mock_get_running_loop)

    # The call we are testing
    ret_val = sentry_task_factory(mock_loop, mock_coro, context=mock_context)

    assert MockTask.called
    assert ret_val == MockTask.return_value

    task_args, task_kwargs = MockTask.call_args
    assert len(task_args) == 1

    coro_param, *_ = task_args
    assert inspect.iscoroutine(coro_param)

    assert "loop" in task_kwargs
    assert task_kwargs["loop"] == mock_loop
    assert "context" in task_kwargs
    assert task_kwargs["context"] == mock_context


@minimum_python_311
@patch("asyncio.get_running_loop")
def test_sentry_task_factory_context_with_factory(mock_get_running_loop):
    mock_loop = mock_get_running_loop.return_value
    mock_coro = MagicMock()
    mock_context = MagicMock()

    # The original task factory will be mocked out here, let's retrieve the value for later
    orig_task_factory = mock_loop.get_task_factory.return_value
    orig_task_factory._is_sentry_task_factory = False

    # Retieve sentry task factory (since it is an inner function within patch_asyncio)
    sentry_task_factory = get_sentry_task_factory(mock_get_running_loop)

    # The call we are testing
    ret_val = sentry_task_factory(mock_loop, mock_coro, context=mock_context)

    assert orig_task_factory.called
    assert ret_val == orig_task_factory.return_value

    task_factory_args, task_factory_kwargs = orig_task_factory.call_args
    assert len(task_factory_args) == 2

    loop_arg, coro_arg = task_factory_args
    assert loop_arg == mock_loop
    assert inspect.iscoroutine(coro_arg)

    assert "context" in task_factory_kwargs
    assert task_factory_kwargs["context"] == mock_context


@minimum_python_38
@pytest.mark.asyncio(loop_scope="module")
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        integrations=[AsyncioIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="something"):
            tasks = [
                asyncio.create_task(foo()),
            ]
            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="something"):
            tasks = [
                asyncio.create_task(foo()),
            ]
            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    sentry_sdk.flush()

    if span_streaming:
        segment = items.pop().payload
        assert segment["is_segment"] is True
        assert segment["attributes"]["sentry.origin"] == "manual"

        spans = [item.payload for item in items]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.origin"] == "auto.function.asyncio"
    else:
        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.function.asyncio"


@minimum_python_38
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_task_spans_false(
    sentry_init,
    capture_events,
    capture_items,
    uninstall_integration,
    span_streaming,
):
    uninstall_integration("asyncio")

    sentry_init(
        traces_sample_rate=1.0,
        integrations=[
            AsyncioIntegration(task_spans=False),
        ],
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_no_spans"):
            tasks = [asyncio.create_task(foo()), asyncio.create_task(bar())]
            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_no_spans"):
            tasks = [asyncio.create_task(foo()), asyncio.create_task(bar())]
            await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)

    sentry_sdk.flush()

    if span_streaming:
        segment = items.pop().payload
        assert segment["is_segment"] is True
        assert segment["name"] == "test_no_spans"

        spans = [item.payload for item in items]
        assert len(spans) == 0
    else:
        (transaction_event,) = events

        assert not transaction_event["spans"]


@minimum_python_38
@pytest.mark.asyncio
async def test_enable_asyncio_integration_with_task_spans_false(
    sentry_init,
    capture_events,
    uninstall_integration,
):
    """
    Test that enable_asyncio_integration() helper works with task_spans=False.
    """
    uninstall_integration("asyncio")

    sentry_init(traces_sample_rate=1.0)

    assert "asyncio" not in sentry_sdk.get_client().integrations

    enable_asyncio_integration(task_spans=False)

    assert "asyncio" in sentry_sdk.get_client().integrations
    assert sentry_sdk.get_client().integrations["asyncio"].task_spans is False

    events = capture_events()

    with sentry_sdk.start_transaction(name="test"):
        await asyncio.create_task(foo())

    sentry_sdk.flush()

    (transaction,) = events
    assert not transaction["spans"]


@minimum_python_38
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_delayed_enable_integration(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    assert "asyncio" not in sentry_sdk.get_client().integrations

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test"):
            await asyncio.create_task(foo())

        sentry_sdk.flush()

        assert len(items) == 1
        assert items[0].payload.get("is_segment") is True

        items.clear()

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            await asyncio.create_task(foo())

        assert len(events) == 1
        (transaction,) = events
        assert not transaction["spans"]

    enable_asyncio_integration()

    assert "asyncio" in sentry_sdk.get_client().integrations

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test"):
            await asyncio.create_task(foo())

        sentry_sdk.flush()

        segment = items.pop().payload
        assert segment["is_segment"] is True

        spans = [item.payload for item in items]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.origin"] == "auto.function.asyncio"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            await asyncio.create_task(foo())

        assert len(events) == 1
        (transaction,) = events
        assert transaction["spans"]
        assert transaction["spans"][0]["origin"] == "auto.function.asyncio"


@minimum_python_38
@pytest.mark.asyncio
async def test_delayed_enable_integration_with_options(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)

    assert "asyncio" not in sentry_sdk.get_client().integrations

    mock_init = MagicMock(return_value=None)
    mock_setup_once = MagicMock()
    with patch(
        "sentry_sdk.integrations.asyncio.AsyncioIntegration.__init__", mock_init
    ):
        with patch(
            "sentry_sdk.integrations.asyncio.AsyncioIntegration.setup_once",
            mock_setup_once,
        ):
            enable_asyncio_integration("arg", kwarg="kwarg")

    assert "asyncio" in sentry_sdk.get_client().integrations
    mock_init.assert_called_once_with("arg", kwarg="kwarg")
    mock_setup_once.assert_called_once()


@minimum_python_38
@pytest.mark.asyncio
async def test_delayed_enable_enabled_integration(sentry_init, uninstall_integration):
    # Ensure asyncio integration is not already installed from previous tests
    uninstall_integration("asyncio")

    integration = AsyncioIntegration()
    sentry_init(integrations=[integration], traces_sample_rate=1.0)

    assert "asyncio" in sentry_sdk.get_client().integrations

    # Get the task factory after initial setup - it should be Sentry's
    loop = asyncio.get_running_loop()
    task_factory_before = loop.get_task_factory()
    assert getattr(task_factory_before, "_is_sentry_task_factory", False) is True

    enable_asyncio_integration()

    assert "asyncio" in sentry_sdk.get_client().integrations

    # The task factory should be the same (loop not re-patched)
    task_factory_after = loop.get_task_factory()
    assert task_factory_before is task_factory_after


@minimum_python_38
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_delayed_enable_integration_after_disabling(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        disabled_integrations=[AsyncioIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    assert "asyncio" not in sentry_sdk.get_client().integrations

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test"):
            await asyncio.create_task(foo())

        sentry_sdk.flush()

        assert len(items) == 1
        assert items[0].payload.get("is_segment") is True

        items.clear()

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            await asyncio.create_task(foo())

        assert len(events) == 1
        (transaction,) = events
        assert not transaction["spans"]

    enable_asyncio_integration()

    assert "asyncio" in sentry_sdk.get_client().integrations

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test"):
            await asyncio.create_task(foo())

        sentry_sdk.flush()

        segment = items.pop().payload
        assert segment["is_segment"] is True

        spans = [item.payload for item in items]
        assert len(spans) == 1
        assert spans[0]["attributes"]["sentry.origin"] == "auto.function.asyncio"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test"):
            await asyncio.create_task(foo())

        assert len(events) == 1
        (transaction,) = events
        assert transaction["spans"]
        assert transaction["spans"][0]["origin"] == "auto.function.asyncio"


@minimum_python_39
@pytest.mark.asyncio(loop_scope="module")
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_internal_tasks_not_wrapped(
    sentry_init, capture_events, capture_items, span_streaming
):
    from sentry_sdk.utils import mark_sentry_task_internal

    sentry_init(
        integrations=[AsyncioIntegration()],
        traces_sample_rate=1.0,
        _experiments={
            "trace_lifecycle": "stream" if span_streaming else "static",
        },
    )

    async def user_task():
        await asyncio.sleep(0.01)
        return "user_result"

    async def internal_task():
        await asyncio.sleep(0.01)
        return "internal_result"

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="test_streamed_span"):
            user_task_obj = asyncio.create_task(user_task())

            with mark_sentry_task_internal():
                internal_task_obj = asyncio.create_task(internal_task())

            user_result = await user_task_obj
            internal_result = await internal_task_obj

    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction"):
            user_task_obj = asyncio.create_task(user_task())

            with mark_sentry_task_internal():
                internal_task_obj = asyncio.create_task(internal_task())

            user_result = await user_task_obj
            internal_result = await internal_task_obj

    assert user_result == "user_result"
    assert internal_result == "internal_result"

    sentry_sdk.flush()

    if span_streaming:
        assert len(items) == 2

        segment = items.pop().payload
        assert segment["is_segment"] is True
        assert segment["name"] == "test_streamed_span"

        spans = [item.payload for item in items]
        assert len(spans) == 1
        assert spans[0]["name"].endswith("user_task")
    else:
        assert len(events) == 1
        transaction = events[0]

        user_spans = []
        internal_spans = []

        for span in transaction.get("spans", []):
            if "user_task" in span.get("description", ""):
                user_spans.append(span)
            elif "internal_task" in span.get("description", ""):
                internal_spans.append(span)

        assert len(user_spans) > 0, (
            f"User task should have been traced. All spans: {[s.get('description') for s in transaction.get('spans', [])]}"
        )
        assert len(internal_spans) == 0, (
            f"Internal task should NOT have been traced. All spans: {[s.get('description') for s in transaction.get('spans', [])]}"
        )


@minimum_python_38
def test_loop_close_patching(sentry_init):
    sentry_init(integrations=[AsyncioIntegration()])

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        with patch("asyncio.get_running_loop", return_value=loop):
            assert not hasattr(loop, "_sentry_flush_patched")
            AsyncioIntegration.setup_once()
            assert hasattr(loop, "_sentry_flush_patched")
            assert loop._sentry_flush_patched is True

    finally:
        if not loop.is_closed():
            loop.close()


@minimum_python_38
def test_loop_close_flushes_async_transport(sentry_init):
    from sentry_sdk.transport import ASYNC_TRANSPORT_AVAILABLE, AsyncHttpTransport

    if not ASYNC_TRANSPORT_AVAILABLE:
        pytest.skip("httpcore[asyncio] not installed")

    sentry_init(integrations=[AsyncioIntegration()])

    # Save the current event loop to restore it later
    try:
        original_loop = asyncio.get_event_loop()
    except RuntimeError:
        original_loop = None

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        with patch("asyncio.get_running_loop", return_value=loop):
            AsyncioIntegration.setup_once()

        mock_client = Mock()
        mock_transport = Mock(spec=AsyncHttpTransport)
        mock_client.transport = mock_transport
        mock_client.close_async = AsyncMock(return_value=None)

        with patch("sentry_sdk.get_client", return_value=mock_client):
            loop.close()

        mock_client.close_async.assert_called_once()
        mock_client.close_async.assert_awaited_once()

    finally:
        if not loop.is_closed():
            loop.close()
        if original_loop:
            asyncio.set_event_loop(original_loop)



import sentry_sdk


def test_full_stack_frames_default(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    def foo():
        try:
            bar()
        except Exception as e:
            sentry_sdk.capture_exception(e)

    def bar():
        raise Exception("This is a test exception")

    foo()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]

    assert len(frames) == 2
    assert frames[-1]["function"] == "bar"
    assert frames[-2]["function"] == "foo"


def test_full_stack_frames_enabled(sentry_init, capture_events):
    sentry_init(
        add_full_stack=True,
    )
    events = capture_events()

    def foo():
        try:
            bar()
        except Exception as e:
            sentry_sdk.capture_exception(e)

    def bar():
        raise Exception("This is a test exception")

    foo()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]

    assert len(frames) > 2
    assert frames[-1]["function"] == "bar"
    assert frames[-2]["function"] == "foo"
    assert frames[-3]["function"] == "foo"
    assert frames[-4]["function"] == "test_full_stack_frames_enabled"


def test_full_stack_frames_enabled_truncated(sentry_init, capture_events):
    sentry_init(
        add_full_stack=True,
        max_stack_frames=3,
    )
    events = capture_events()

    def foo():
        try:
            bar()
        except Exception as e:
            sentry_sdk.capture_exception(e)

    def bar():
        raise Exception("This is a test exception")

    foo()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]

    assert len(frames) == 3
    assert frames[-1]["function"] == "bar"
    assert frames[-2]["function"] == "foo"
    assert frames[-3]["function"] == "foo"


def test_full_stack_frames_default_no_truncation_happening(sentry_init, capture_events):
    sentry_init(
        max_stack_frames=1,  # this is ignored if add_full_stack=False (which is the default)
    )
    events = capture_events()

    def foo():
        try:
            bar()
        except Exception as e:
            sentry_sdk.capture_exception(e)

    def bar():
        raise Exception("This is a test exception")

    foo()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]

    assert len(frames) == 2
    assert frames[-1]["function"] == "bar"
    assert frames[-2]["function"] == "foo"
